import asyncio
import functools
import logging
import sys

import pytest
import zigpy.config

from bellows import config, ezsp, uart
from bellows.exception import EzspError, InvalidCommandError
import bellows.ezsp.v4.types as v4_t
import bellows.types as t

if sys.version_info[:2] < (3, 11):
    from async_timeout import timeout as asyncio_timeout  # pragma: no cover
else:
    from asyncio import timeout as asyncio_timeout  # pragma: no cover

from unittest.mock import ANY, AsyncMock, MagicMock, call, patch, sentinel

from bellows.ezsp.v9.commands import GetTokenDataRsp

DEVICE_CONFIG = {
    zigpy.config.CONF_DEVICE_PATH: "/dev/null",
    zigpy.config.CONF_DEVICE_BAUDRATE: 115200,
}


@pytest.fixture
async def ezsp_f():
    api = ezsp.EZSP(DEVICE_CONFIG)
    gw = MagicMock(spec_set=uart.Gateway)
    with patch("bellows.uart.connect", new=AsyncMock(return_value=gw)):
        await api.connect()
        yield api


async def make_ezsp(version=4) -> ezsp.EZSP:
    api = ezsp.EZSP(DEVICE_CONFIG)
    gw = MagicMock(spec_set=uart.Gateway)

    with patch("bellows.uart.connect", new=AsyncMock(return_value=gw)):
        await api.connect()

    assert api._ezsp_version == 4

    with patch.object(api, "_command", new=AsyncMock(return_value=[version, 0, 0])):
        await api.version()

    assert api._ezsp_version == version

    return api


async def test_connect(ezsp_f, monkeypatch):
    connected = False

    async def mockconnect(*args, **kwargs):
        nonlocal connected
        connected = True

    monkeypatch.setattr(uart, "connect", mockconnect)
    ezsp_f._gw = None

    await ezsp_f.connect()
    assert connected


async def test_reset(ezsp_f):
    ezsp_f.stop_ezsp = MagicMock()
    ezsp_f.start_ezsp = MagicMock()
    reset_mock = AsyncMock()
    ezsp_f._gw.reset = MagicMock(side_effect=reset_mock)

    await ezsp_f.reset()
    assert ezsp_f._gw.reset.call_count == 1
    assert ezsp_f.start_ezsp.call_count == 1
    assert ezsp_f.stop_ezsp.call_count == 1
    assert len(ezsp_f._callbacks) == 1


def test_close(ezsp_f):
    closed = False

    def close_mock(*args):
        nonlocal closed
        closed = True

    ezsp_f._gw.close = close_mock
    ezsp_f.close()
    assert closed is True
    assert ezsp_f._gw is None


def test_attr(ezsp_f):
    m = ezsp_f.getValue
    assert isinstance(m, functools.partial)
    assert callable(m)


async def test_non_existent_attr(ezsp_f):
    with pytest.raises(AttributeError):
        await ezsp_f.nonexistentMethod()


async def test_command(ezsp_f):
    ezsp_f.start_ezsp()
    with patch.object(ezsp_f._protocol, "command") as cmd_mock:
        await ezsp_f.nop()
    assert cmd_mock.call_count == 1


async def test_command_ezsp_stopped(ezsp_f):
    with pytest.raises(EzspError):
        await ezsp_f._command("version")


async def _test_list_command(ezsp_f, mockcommand):
    ezsp_f._command = mockcommand
    return await ezsp_f._list_command(
        "startScan", ["networkFoundHandler"], "scanCompleteHandler", 1
    )


async def test_list_command(ezsp_f):
    async def mockcommand(name, *args):
        assert name == "startScan"
        ezsp_f.frame_received(b"\x01\x00\x1b" + b"\x00" * 20)
        ezsp_f.frame_received(b"\x02\x00\x1b" + b"\x00" * 20)
        ezsp_f.frame_received(b"\x03\x00\x1c" + b"\x00" * 20)

        return [0]

    result = await _test_list_command(ezsp_f, mockcommand)
    assert len(result) == 2


async def test_list_command_initial_failure(ezsp_f):
    async def mockcommand(name, *args):
        assert name == "startScan"
        return [1]

    with pytest.raises(Exception):
        await _test_list_command(ezsp_f, mockcommand)


async def test_list_command_later_failure(ezsp_f):
    async def mockcommand(name, *args):
        assert name == "startScan"
        ezsp_f.frame_received(b"\x01\x00\x1b" + b"\x00" * 20)
        ezsp_f.frame_received(b"\x02\x00\x1b" + b"\x00" * 20)
        ezsp_f.frame_received(b"\x03\x00\x1c\x01\x01")

        return [0]

    with pytest.raises(Exception):
        await _test_list_command(ezsp_f, mockcommand)


async def _test_form_network(ezsp_f, initial_result, final_result):
    async def mockcommand(name, *args):
        assert name == "formNetwork"
        ezsp_f.frame_received(b"\x01\x00\x19" + final_result)
        return initial_result

    ezsp_f._command = mockcommand

    await ezsp_f.formNetwork(MagicMock())


async def test_form_network(ezsp_f):
    await _test_form_network(ezsp_f, [0], b"\x90")


async def test_form_network_fail(ezsp_f):
    with pytest.raises(Exception):
        await _test_form_network(ezsp_f, [1], b"\x90")


async def test_form_network_fail_stack_status(ezsp_f):
    with pytest.raises(Exception):
        await _test_form_network(ezsp_f, [0], b"\x00")


def test_receive_new(ezsp_f):
    callback = MagicMock()
    ezsp_f.add_callback(callback)
    ezsp_f.frame_received(b"\x00\xff\x00\x04\x05\x06\x00")
    assert callback.call_count == 1


def test_callback(ezsp_f):
    testcb = MagicMock()

    cbid = ezsp_f.add_callback(testcb)
    ezsp_f.handle_callback(1, 2, 3)

    assert testcb.call_count == 1

    ezsp_f.remove_callback(cbid)
    ezsp_f.handle_callback(4, 5, 6)
    assert testcb.call_count == 1


def test_callback_multi(ezsp_f):
    testcb = MagicMock()

    cbid1 = ezsp_f.add_callback(testcb)
    ezsp_f.add_callback(testcb)

    ezsp_f.handle_callback(1, 2, 3)

    assert testcb.call_count == 2

    ezsp_f.remove_callback(cbid1)

    ezsp_f.handle_callback(4, 5, 6)
    testcb.assert_has_calls([call(1, 2, 3), call(1, 2, 3), call(4, 5, 6)])


def test_callback_exc(ezsp_f):
    testcb = MagicMock()
    testcb.side_effect = Exception("Testing")

    ezsp_f.add_callback(testcb)
    ezsp_f.handle_callback(1)
    assert testcb.call_count == 1


@pytest.mark.parametrize("version, call_count", ((4, 1), (5, 2), (6, 2), (99, 2)))
async def test_change_version(ezsp_f, version, call_count):
    def mockcommand(name, *args):
        assert name == "version"
        ezsp_f.frame_received(b"\x01\x00\x00\x21\x22\x23\x24")
        fut = asyncio.Future()
        fut.set_result([version, 2, 2046])
        return fut

    ezsp_f._command = MagicMock(side_effect=mockcommand)
    await ezsp_f.version()
    assert ezsp_f.ezsp_version == version
    assert ezsp_f._command.call_count == call_count


def test_stop_ezsp(ezsp_f):
    ezsp_f._ezsp_event.set()
    ezsp_f.stop_ezsp()
    assert ezsp_f._ezsp_event.is_set() is False


def test_start_ezsp(ezsp_f):
    ezsp_f._ezsp_event.clear()
    ezsp_f.start_ezsp()
    assert ezsp_f._ezsp_event.is_set() is True


def test_connection_lost(ezsp_f):
    ezsp_f.enter_failed_state = MagicMock(spec_set=ezsp_f.enter_failed_state)
    ezsp_f.connection_lost(sentinel.exc)
    assert ezsp_f.enter_failed_state.call_count == 1


async def test_enter_failed_state(ezsp_f):
    ezsp_f.stop_ezsp = MagicMock(spec_set=ezsp_f.stop_ezsp)
    cb = MagicMock(spec_set=ezsp_f.handle_callback)
    ezsp_f.add_callback(cb)
    ezsp_f.enter_failed_state(sentinel.error)
    await asyncio.sleep(0)
    assert ezsp_f.stop_ezsp.call_count == 1
    assert cb.call_count == 1
    assert cb.call_args[0][1][0] == sentinel.error


async def test_no_close_without_callback(ezsp_f):
    ezsp_f.stop_ezsp = MagicMock(spec_set=ezsp_f.stop_ezsp)
    ezsp_f.close = MagicMock(spec_set=ezsp_f.close)
    ezsp_f.enter_failed_state(sentinel.error)
    await asyncio.sleep(0)
    assert ezsp_f.stop_ezsp.call_count == 0
    assert ezsp_f.close.call_count == 0


@patch.object(ezsp.EZSP, "version", new_callable=AsyncMock)
@patch.object(ezsp.EZSP, "reset", new_callable=AsyncMock)
@patch("bellows.uart.connect", return_value=MagicMock(spec_set=uart.Gateway))
async def test_ezsp_init(conn_mock, reset_mock, version_mock):
    """Test initialize method."""
    zigpy_config = config.CONFIG_SCHEMA({"device": DEVICE_CONFIG})
    await ezsp.EZSP.initialize(zigpy_config)
    assert conn_mock.await_count == 1
    assert reset_mock.await_count == 1
    assert version_mock.await_count == 1


@patch.object(ezsp.EZSP, "version", side_effect=RuntimeError("Uh oh"))
@patch.object(ezsp.EZSP, "reset", new_callable=AsyncMock)
@patch.object(ezsp.EZSP, "close", new_callable=MagicMock)
@patch("bellows.uart.connect", return_value=MagicMock(spec_set=uart.Gateway))
async def test_ezsp_init_failure(conn_mock, close_mock, reset_mock, version_mock):
    """Test initialize method failing."""
    zigpy_config = config.CONFIG_SCHEMA({"device": DEVICE_CONFIG})

    with pytest.raises(RuntimeError):
        await ezsp.EZSP.initialize(zigpy_config)

    assert conn_mock.await_count == 1
    assert reset_mock.await_count == 1
    assert version_mock.await_count == 1
    assert close_mock.call_count == 1


async def test_ezsp_newer_version(ezsp_f):
    """Test newer version of ezsp."""
    with patch.object(
        ezsp_f, "_command", new=AsyncMock(return_value=(9, 0x12, 0x12345))
    ):
        await ezsp_f.version()


async def test_board_info(ezsp_f):
    """Test getting board info."""

    def cmd_mock(config):
        async def replacement(*args):
            return tuple(config[args])

        return replacement

    with patch.object(
        ezsp_f,
        "_command",
        new=cmd_mock(
            {
                ("getMfgToken", t.EzspMfgTokenId.MFG_BOARD_NAME): (
                    b"\xfe\xff\xff\xff",
                ),
                ("getMfgToken", t.EzspMfgTokenId.MFG_STRING): (
                    b"Manufacturer\xff\xff\xff",
                ),
                ("getValue", ezsp_f.types.EzspValueId.VALUE_VERSION_INFO): (
                    0x00,
                    b"\x01\x02\x03\x04\x05\x06",
                ),
            }
        ),
    ):
        mfg, brd, ver = await ezsp_f.get_board_info()

    assert mfg == "Manufacturer"
    assert brd == "0xFE"
    assert ver == "3.4.5.6 build 513"

    with patch.object(
        ezsp_f,
        "_command",
        new=cmd_mock(
            {
                ("getMfgToken", t.EzspMfgTokenId.MFG_BOARD_NAME): (
                    b"\xfe\xff\xff\xff",
                ),
                ("getMfgToken", t.EzspMfgTokenId.MFG_STRING): (
                    b"Manufacturer\xff\xff\xff",
                ),
                ("getValue", ezsp_f.types.EzspValueId.VALUE_VERSION_INFO): (
                    0x01,
                    b"\x01\x02\x03\x04\x05\x06",
                ),
            }
        ),
    ):
        mfg, brd, ver = await ezsp_f.get_board_info()

    assert mfg == "Manufacturer"
    assert brd == "0xFE"
    assert ver is None

    with patch.object(
        ezsp_f,
        "_command",
        new=cmd_mock(
            {
                ("getMfgToken", t.EzspMfgTokenId.MFG_BOARD_NAME): (
                    b"SkyBlue v0.1\x00\xff\xff\xff",
                ),
                ("getMfgToken", t.EzspMfgTokenId.MFG_STRING): (
                    b"Nabu Casa\x00\xff\xff\xff\xff\xff\xff",
                ),
                ("getValue", ezsp_f.types.EzspValueId.VALUE_VERSION_INFO): (
                    0x00,
                    b"\xbf\x00\x07\x01\x00\x00\xaa",
                ),
            }
        ),
    ):
        mfg, brd, ver = await ezsp_f.get_board_info()

    assert mfg == "Nabu Casa"
    assert brd == "SkyBlue v0.1"
    assert ver == "7.1.0.0 build 191"

    with patch.object(
        ezsp_f,
        "_command",
        new=cmd_mock(
            {
                ("getMfgToken", t.EzspMfgTokenId.MFG_BOARD_NAME): (b"\xff" * 16,),
                ("getMfgToken", t.EzspMfgTokenId.MFG_STRING): (b"\xff" * 16,),
                ("getValue", ezsp_f.types.EzspValueId.VALUE_VERSION_INFO): (
                    0x00,
                    b"\xbf\x00\x07\x01\x00\x00\xaa",
                ),
            }
        ),
    ):
        mfg, brd, ver = await ezsp_f.get_board_info()

    assert mfg is None
    assert brd is None
    assert ver == "7.1.0.0 build 191"


async def test_pre_permit(ezsp_f):
    with patch("bellows.ezsp.v4.EZSPv4.pre_permit") as pre_mock:
        await ezsp_f.pre_permit(sentinel.time)
        assert pre_mock.call_count == 1
        assert pre_mock.await_count == 1


async def test_update_policies(ezsp_f):
    with patch("bellows.ezsp.v4.EZSPv4.update_policies") as pol_mock:
        await ezsp_f.update_policies(sentinel.time)
        assert pol_mock.call_count == 1
        assert pol_mock.await_count == 1


async def test_set_source_routing_set_concentrator(ezsp_f):
    """Test enabling source routing."""
    with patch.object(ezsp_f, "setConcentrator", new=AsyncMock()) as cnc_mock:
        cnc_mock.return_value = (ezsp_f.types.EmberStatus.SUCCESS,)
        await ezsp_f.set_source_routing()
        assert cnc_mock.await_count == 1

        cnc_mock.return_value = (ezsp_f.types.EmberStatus.ERR_FATAL,)
        await ezsp_f.set_source_routing()
        assert cnc_mock.await_count == 2


async def test_set_source_routing_ezsp_v8(ezsp_f):
    """Test enabling source routing on EZSPv8."""

    ezsp_f._ezsp_version = 8
    ezsp_f.setConcentrator = AsyncMock(return_value=(ezsp_f.types.EmberStatus.SUCCESS,))
    ezsp_f.setSourceRouteDiscoveryMode = AsyncMock()

    await ezsp_f.set_source_routing()
    assert len(ezsp_f.setSourceRouteDiscoveryMode.mock_calls) == 1


async def test_leave_network_error(ezsp_f):
    """Test EZSP leaveNetwork command failure."""

    with patch.object(ezsp_f, "_command", new_callable=AsyncMock) as cmd_mock:
        cmd_mock.return_value = [t.EmberStatus.ERR_FATAL]
        with pytest.raises(EzspError):
            await ezsp_f.leaveNetwork(timeout=0.01)


async def test_leave_network_no_stack_status(ezsp_f):
    """Test EZSP leaveNetwork command, no stackStatusHandler callback."""

    with patch.object(ezsp_f, "_command", new_callable=AsyncMock) as cmd_mock:
        cmd_mock.return_value = [t.EmberStatus.SUCCESS]
        with pytest.raises(asyncio.TimeoutError):
            await ezsp_f.leaveNetwork(timeout=0.01)


async def test_leave_network(ezsp_f):
    """Test EZSP leaveNetwork command."""

    async def _mock_cmd(*args, **kwargs):
        ezsp_f.handle_callback("stackStatusHandler", [t.EmberStatus.NETWORK_UP])
        ezsp_f.handle_callback("stackStatusHandler", [t.EmberStatus.NETWORK_UP])
        ezsp_f.handle_callback("stackStatusHandler", [t.EmberStatus.NETWORK_DOWN])
        return [t.EmberStatus.SUCCESS]

    with patch.object(ezsp_f, "_command", new_callable=AsyncMock) as cmd_mock:
        cmd_mock.side_effect = _mock_cmd
        await ezsp_f.leaveNetwork(timeout=0.01)


@pytest.mark.parametrize(
    "value, expected_result",
    [
        (b"\xFF" * 8, True),
        (bytes.fromhex("0846b8a11c004b1200"), False),
        (b"", False),
    ],
)
async def test_can_burn_userdata_custom_eui64(ezsp_f, value, expected_result):
    """Test detecting if a custom EUI64 has been written."""
    ezsp_f.getMfgToken = AsyncMock(return_value=[value])

    assert await ezsp_f.can_burn_userdata_custom_eui64() == expected_result

    ezsp_f.getMfgToken.assert_called_once_with(t.EzspMfgTokenId.MFG_CUSTOM_EUI_64)


@pytest.mark.parametrize(
    "tokens, expected_key, expected_result",
    [
        ({}, None, False),
        (
            {t.NV3KeyId.CREATOR_STACK_RESTORED_EUI64: b"\xAA" * 8},
            t.NV3KeyId.CREATOR_STACK_RESTORED_EUI64,
            True,
        ),
        (
            {t.NV3KeyId.NVM3KEY_STACK_RESTORED_EUI64: b"\xAA" * 8},
            t.NV3KeyId.NVM3KEY_STACK_RESTORED_EUI64,
            True,
        ),
    ],
)
async def test_can_rewrite_custom_eui64(ezsp_f, tokens, expected_key, expected_result):
    """Test detecting if a custom EUI64 can be rewritten in NV3."""

    def get_token_data(key, index):
        if key not in tokens or index != 0:
            return GetTokenDataRsp(status=t.EmberStatus.ERR_FATAL)

        return GetTokenDataRsp(status=t.EmberStatus.SUCCESS, value=tokens[key])

    ezsp_f.getTokenData = AsyncMock(side_effect=get_token_data)

    key = await ezsp_f._get_nv3_restored_eui64_key()
    assert key == expected_key
    assert await ezsp_f.can_rewrite_custom_eui64() == expected_result


async def test_can_rewrite_custom_eui64_old_ezsp(ezsp_f):
    """Test detecting if a custom EUI64 can be rewritten in NV3, but with old EZSP."""

    assert await ezsp_f._get_nv3_restored_eui64_key() is None
    assert not await ezsp_f.can_rewrite_custom_eui64()


async def test_write_custom_eui64(ezsp_f):
    """Test writing a custom EUI64."""

    old_eui64 = t.EUI64.convert("AA" * 8)
    new_eui64 = t.EUI64.convert("BB" * 8)

    ezsp_f.getEui64 = AsyncMock(return_value=[old_eui64])
    ezsp_f.setMfgToken = AsyncMock(return_value=[t.EmberStatus.SUCCESS])
    ezsp_f.setTokenData = AsyncMock(return_value=[t.EmberStatus.SUCCESS])
    ezsp_f._get_mfg_custom_eui_64 = AsyncMock(return_value=old_eui64)
    ezsp_f._get_nv3_restored_eui64_key = AsyncMock(return_value=None)

    # Nothing is done when the EUI64 write is a no-op
    await ezsp_f.write_custom_eui64(old_eui64)

    ezsp_f.setMfgToken.assert_not_called()
    ezsp_f.setTokenData.assert_not_called()

    # If NV3 exists, all writes succeed
    ezsp_f._get_nv3_restored_eui64_key.return_value = (
        t.NV3KeyId.NVM3KEY_STACK_RESTORED_EUI64
    )
    await ezsp_f.write_custom_eui64(new_eui64)
    await ezsp_f.write_custom_eui64(new_eui64, burn_into_userdata=True)

    ezsp_f.setMfgToken.assert_not_called()
    ezsp_f.setTokenData.mock_calls == 2 * [
        call(
            t.NV3KeyId.NVM3KEY_STACK_RESTORED_EUI64,
            0,
            new_eui64.serialize(),
        )
    ]

    ezsp_f.setTokenData.reset_mock()

    # If NV3 does not and the MFG token does not, we conditionally write
    ezsp_f._get_mfg_custom_eui_64.return_value = None
    ezsp_f._get_nv3_restored_eui64_key.return_value = None

    with pytest.raises(EzspError):
        await ezsp_f.write_custom_eui64(new_eui64)

    # Burn kwarg not passed, so nothing is done
    ezsp_f.setMfgToken.assert_not_called()
    ezsp_f.setTokenData.assert_not_called()

    await ezsp_f.write_custom_eui64(new_eui64, burn_into_userdata=True)

    ezsp_f.setMfgToken.assert_called_once_with(
        t.EzspMfgTokenId.MFG_CUSTOM_EUI_64, new_eui64.serialize()
    )
    ezsp_f.setTokenData.assert_not_called()

    ezsp_f.setMfgToken.reset_mock()

    # If no method is viable, throw an error
    ezsp_f._get_mfg_custom_eui_64.return_value = old_eui64

    with pytest.raises(EzspError):
        await ezsp_f.write_custom_eui64(new_eui64)

    with pytest.raises(EzspError):
        await ezsp_f.write_custom_eui64(new_eui64, burn_into_userdata=True)

    ezsp_f.setMfgToken.assert_not_called()
    ezsp_f.setTokenData.assert_not_called()


async def test_write_custom_eui64_rcp(ezsp_f):
    """Test writing a custom EUI64 with RPC firmware."""

    old_eui64 = t.EUI64.convert("AA" * 8)
    new_eui64 = t.EUI64.convert("BB" * 8)

    ezsp_f.getEui64 = AsyncMock(return_value=[old_eui64])
    ezsp_f.setMfgToken = AsyncMock(return_value=[t.EmberStatus.INVALID_CALL])
    ezsp_f.setTokenData = AsyncMock(return_value=[t.EmberStatus.SUCCESS])

    # RCP firmware does not support manufacturing tokens
    ezsp_f.getMfgToken = AsyncMock(return_value=[b""])
    ezsp_f.getTokenData = AsyncMock(
        return_value=GetTokenDataRsp(status=t.EmberStatus.SUCCESS, value=b"\xFF" * 8)
    )

    await ezsp_f.write_custom_eui64(new_eui64)

    ezsp_f.setMfgToken.assert_not_called()
    ezsp_f.setTokenData.mock_calls == [
        call(
            t.NV3KeyId.NVM3KEY_STACK_RESTORED_EUI64,
            0,
            new_eui64.serialize(),
        )
    ]


@patch.object(ezsp.EZSP, "version", new_callable=AsyncMock)
@patch.object(ezsp.EZSP, "reset", new_callable=AsyncMock)
@patch("bellows.uart.connect", return_value=MagicMock(spec_set=uart.Gateway))
async def test_ezsp_init_zigbeed(conn_mock, reset_mock, version_mock):
    """Test initialize method with a received startup reset frame."""
    zigpy_config = config.CONFIG_SCHEMA(
        {
            "device": {
                **DEVICE_CONFIG,
                zigpy.config.CONF_DEVICE_PATH: "socket://localhost:1234",
            }
        }
    )

    gw_wait_reset_mock = conn_mock.return_value.wait_for_startup_reset = AsyncMock()

    await ezsp.EZSP.initialize(zigpy_config)

    assert conn_mock.await_count == 1
    assert reset_mock.await_count == 0  # Reset is not called
    assert gw_wait_reset_mock.await_count == 1
    assert version_mock.await_count == 1


@patch.object(ezsp.EZSP, "version", new_callable=AsyncMock)
@patch.object(ezsp.EZSP, "reset", new_callable=AsyncMock)
@patch("bellows.uart.connect", return_value=MagicMock(spec_set=uart.Gateway))
@patch("bellows.ezsp.NETWORK_COORDINATOR_STARTUP_RESET_WAIT", 0.01)
async def test_ezsp_init_zigbeed_timeout(conn_mock, reset_mock, version_mock):
    """Test initialize method with a received startup reset frame."""
    zigpy_config = config.CONFIG_SCHEMA(
        {
            "device": {
                **DEVICE_CONFIG,
                zigpy.config.CONF_DEVICE_PATH: "socket://localhost:1234",
            }
        }
    )

    async def wait_forever(*args, **kwargs):
        return await asyncio.get_running_loop().create_future()

    gw_wait_reset_mock = conn_mock.return_value.wait_for_startup_reset = AsyncMock(
        side_effect=wait_forever
    )

    await ezsp.EZSP.initialize(zigpy_config)

    assert conn_mock.await_count == 1
    assert reset_mock.await_count == 1  # Reset will be called
    assert gw_wait_reset_mock.await_count == 1
    assert version_mock.await_count == 1


async def test_wait_for_stack_status(ezsp_f):
    assert not ezsp_f._stack_status_listeners[t.EmberStatus.NETWORK_DOWN]

    # Cancellation clears handlers
    with ezsp_f.wait_for_stack_status(t.EmberStatus.NETWORK_DOWN) as stack_status:
        with pytest.raises(asyncio.TimeoutError):
            async with asyncio_timeout(0.1):
                assert ezsp_f._stack_status_listeners[t.EmberStatus.NETWORK_DOWN]
                await stack_status

    assert not ezsp_f._stack_status_listeners[t.EmberStatus.NETWORK_DOWN]

    # Receiving multiple also works
    with ezsp_f.wait_for_stack_status(t.EmberStatus.NETWORK_DOWN) as stack_status:
        ezsp_f.handle_callback("stackStatusHandler", [t.EmberStatus.NETWORK_UP])
        ezsp_f.handle_callback("stackStatusHandler", [t.EmberStatus.NETWORK_DOWN])
        ezsp_f.handle_callback("stackStatusHandler", [t.EmberStatus.NETWORK_DOWN])

        await stack_status

    assert not ezsp_f._stack_status_listeners[t.EmberStatus.NETWORK_DOWN]


def test_ezsp_versions(ezsp_f):
    for version in range(4, ezsp.EZSP_LATEST + 1):
        assert version in ezsp_f._BY_VERSION
        assert ezsp_f._BY_VERSION[version].__name__ == f"EZSPv{version}"
        assert ezsp_f._BY_VERSION[version].VERSION == version


async def test_config_initialize_husbzb1(ezsp_f):
    """Test timeouts are properly set for HUSBZB-1."""

    ezsp_f._ezsp_version = 4

    ezsp_f.getConfigurationValue = AsyncMock(return_value=(t.EzspStatus.SUCCESS, 0))
    ezsp_f.setConfigurationValue = AsyncMock(return_value=(t.EzspStatus.SUCCESS,))
    ezsp_f.networkState = AsyncMock(return_value=(t.EmberNetworkStatus.JOINED_NETWORK,))

    expected_calls = [
        call(v4_t.EzspConfigId.CONFIG_SOURCE_ROUTE_TABLE_SIZE, 16),
        call(v4_t.EzspConfigId.CONFIG_END_DEVICE_POLL_TIMEOUT, 60),
        call(v4_t.EzspConfigId.CONFIG_END_DEVICE_POLL_TIMEOUT_SHIFT, 8),
        call(v4_t.EzspConfigId.CONFIG_INDIRECT_TRANSMISSION_TIMEOUT, 7680),
        call(v4_t.EzspConfigId.CONFIG_STACK_PROFILE, 2),
        call(v4_t.EzspConfigId.CONFIG_SUPPORTED_NETWORKS, 1),
        call(v4_t.EzspConfigId.CONFIG_MULTICAST_TABLE_SIZE, 16),
        call(v4_t.EzspConfigId.CONFIG_TRUST_CENTER_ADDRESS_CACHE_SIZE, 2),
        call(v4_t.EzspConfigId.CONFIG_SECURITY_LEVEL, 5),
        call(v4_t.EzspConfigId.CONFIG_ADDRESS_TABLE_SIZE, 16),
        call(v4_t.EzspConfigId.CONFIG_PAN_ID_CONFLICT_REPORT_THRESHOLD, 2),
        call(v4_t.EzspConfigId.CONFIG_KEY_TABLE_SIZE, 4),
        call(v4_t.EzspConfigId.CONFIG_MAX_END_DEVICE_CHILDREN, 32),
        call(
            v4_t.EzspConfigId.CONFIG_APPLICATION_ZDO_FLAGS,
            (
                v4_t.EmberZdoConfigurationFlags.APP_HANDLES_UNSUPPORTED_ZDO_REQUESTS
                | v4_t.EmberZdoConfigurationFlags.APP_RECEIVES_SUPPORTED_ZDO_REQUESTS
            ),
        ),
        call(v4_t.EzspConfigId.CONFIG_PACKET_BUFFER_COUNT, 255),
    ]

    await ezsp_f.write_config({})
    assert ezsp_f.setConfigurationValue.mock_calls == expected_calls


@pytest.mark.parametrize("version", ezsp.EZSP._BY_VERSION)
async def test_config_initialize(version: int, ezsp_f, caplog):
    """Test config initialization for all protocol versions."""

    assert ezsp_f.ezsp_version == 4

    with patch.object(ezsp_f, "_command", AsyncMock(return_value=[version, 2, 2046])):
        await ezsp_f.version()

    assert ezsp_f.ezsp_version == version

    ezsp_f.getConfigurationValue = AsyncMock(return_value=(t.EzspStatus.SUCCESS, 0))
    ezsp_f.setConfigurationValue = AsyncMock(return_value=(t.EzspStatus.SUCCESS,))
    ezsp_f.networkState = AsyncMock(return_value=(t.EmberNetworkStatus.JOINED_NETWORK,))

    ezsp_f.setValue = AsyncMock(return_value=(t.EzspStatus.SUCCESS,))
    ezsp_f.getValue = AsyncMock(return_value=(t.EzspStatus.SUCCESS, b"\xFF"))

    await ezsp_f.write_config({})

    with caplog.at_level(logging.DEBUG):
        ezsp_f.setConfigurationValue.return_value = (t.EzspStatus.ERROR_OUT_OF_MEMORY,)
        await ezsp_f.write_config({})

    assert "Could not set config" in caplog.text
    ezsp_f.setConfigurationValue.return_value = (t.EzspStatus.SUCCESS,)
    caplog.clear()

    # EZSPv6 does not set any values on startup
    if version < 7:
        return

    ezsp_f.setValue.reset_mock()
    ezsp_f.getValue.return_value = (t.EzspStatus.ERROR_INVALID_ID, b"")
    await ezsp_f.write_config({})
    assert len(ezsp_f.setValue.mock_calls) == 1

    ezsp_f.getValue = AsyncMock(return_value=(t.EzspStatus.SUCCESS, b"\xFF"))
    caplog.clear()

    with caplog.at_level(logging.DEBUG):
        ezsp_f.setValue.return_value = (t.EzspStatus.ERROR_INVALID_ID,)
        await ezsp_f.write_config({})

    assert "Could not set value" in caplog.text
    ezsp_f.setValue.return_value = (t.EzspStatus.SUCCESS,)
    caplog.clear()


async def test_cfg_initialize_skip(ezsp_f):
    """Test initialization."""

    ezsp_f.networkState = AsyncMock(return_value=(t.EmberNetworkStatus.JOINED_NETWORK,))

    p1 = patch.object(
        ezsp_f,
        "setConfigurationValue",
        new=AsyncMock(return_value=(t.EzspStatus.SUCCESS,)),
    )
    p2 = patch.object(
        ezsp_f,
        "getConfigurationValue",
        new=AsyncMock(return_value=(t.EzspStatus.SUCCESS, 22)),
    )
    with p1, p2:
        await ezsp_f.write_config({"CONFIG_END_DEVICE_POLL_TIMEOUT": None})

        # Config not set when it is explicitly disabled
        with pytest.raises(AssertionError):
            ezsp_f.setConfigurationValue.assert_called_with(
                v4_t.EzspConfigId.CONFIG_END_DEVICE_POLL_TIMEOUT, ANY
            )

    with p1, p2:
        await ezsp_f.write_config({"CONFIG_MULTICAST_TABLE_SIZE": 123})

        # Config is overridden
        ezsp_f.setConfigurationValue.assert_any_call(
            v4_t.EzspConfigId.CONFIG_MULTICAST_TABLE_SIZE, 123
        )

    with p1, p2:
        await ezsp_f.write_config({})

        # Config is set by default
        ezsp_f.setConfigurationValue.assert_any_call(
            v4_t.EzspConfigId.CONFIG_END_DEVICE_POLL_TIMEOUT, ANY
        )


async def test_reset_custom_eui64(ezsp_f):
    """Test resetting custom EUI64."""
    # No NV3 interface
    ezsp_f.getTokenData = AsyncMock(side_effect=InvalidCommandError)
    ezsp_f.setTokenData = AsyncMock(return_value=[t.EmberStatus.SUCCESS])
    await ezsp_f.reset_custom_eui64()

    assert len(ezsp_f.setTokenData.mock_calls) == 0

    # With NV3 interface
    ezsp_f.getTokenData = AsyncMock(
        return_value=GetTokenDataRsp(status=t.EmberStatus.SUCCESS, value=b"\xAB" * 8)
    )
    await ezsp_f.reset_custom_eui64()
    assert ezsp_f.setTokenData.mock_calls == [
        call(t.NV3KeyId.CREATOR_STACK_RESTORED_EUI64, 0, t.LVBytes32(b"\xFF" * 8))
    ]


def test_empty_frame_received(ezsp_f):
    """Test dropping of invalid, empty frames."""
    ezsp_f._protocol = MagicMock(spec_set=ezsp_f._protocol)
    ezsp_f._protocol.__call__ = MagicMock()
    ezsp_f.frame_received(b"")

    assert ezsp_f._protocol.__call__.mock_calls == []


def test_frame_parsing_error_doesnt_disconnect(ezsp_f, caplog):
    """Test that frame parsing error doesn't trigger a disconnect."""
    ezsp_f._protocol = MagicMock(spec_set=ezsp_f._protocol, side_effect=RuntimeError())

    with caplog.at_level(logging.WARNING):
        ezsp_f.frame_received(b"test")

    assert "Failed to parse frame" in caplog.text



import logging

import pytest
import zigpy.state
import zigpy.types as t
import zigpy.zdo.types as zdo_t

import bellows.types as bellows_t
import bellows.zigbee.util as util

from tests.test_application import ezsp_mock, ieee
from tests.test_application_network_state import network_info, node_info


@pytest.fixture
def zigpy_key(network_info, node_info):
    return network_info.network_key.replace(
        rx_counter=1234567,
        partner_ieee=node_info.ieee,
    )


@pytest.fixture
def ezsp_key(ezsp_mock, network_info, node_info, zigpy_key):
    ezsp = ezsp_mock
    return ezsp.types.EmberKeyStruct(
        bitmask=(
            ezsp.types.EmberKeyStructBitmask.KEY_HAS_SEQUENCE_NUMBER
            | ezsp.types.EmberKeyStructBitmask.KEY_HAS_OUTGOING_FRAME_COUNTER
            | ezsp.types.EmberKeyStructBitmask.KEY_HAS_INCOMING_FRAME_COUNTER
            | ezsp.types.EmberKeyStructBitmask.KEY_HAS_PARTNER_EUI64
        ),
        key=ezsp.types.KeyData(network_info.network_key.key),
        sequenceNumber=zigpy_key.seq,
        outgoingFrameCounter=zigpy_key.tx_counter,
        incomingFrameCounter=zigpy_key.rx_counter,
        partnerEUI64=bellows_t.EUI64(node_info.ieee),
    )


def test_zha_security_normal(network_info, node_info):
    security = util.zha_security(network_info=network_info, use_hashed_tclk=True)

    assert security.preconfiguredTrustCenterEui64 == node_info.ieee
    assert (
        bellows_t.EmberInitialSecurityBitmask.HAVE_TRUST_CENTER_EUI64
        in security.bitmask
    )

    assert (
        security.preconfiguredKey.serialize().hex()
        == network_info.stack_specific["ezsp"]["hashed_tclk"]
    )
    assert (
        bellows_t.EmberInitialSecurityBitmask.TRUST_CENTER_USES_HASHED_LINK_KEY
        in security.bitmask
    )


def test_zha_security_router_unknown_tclk_partner_ieee(network_info):
    security = util.zha_security(
        network_info=network_info.replace(
            tc_link_key=network_info.tc_link_key.replace(partner_ieee=t.EUI64.UNKNOWN)
        ),
        use_hashed_tclk=False,
    )

    # Not set, since we don't know it
    assert security.preconfiguredTrustCenterEui64 == bellows_t.EUI64([0x00] * 8)
    assert (
        bellows_t.EmberInitialSecurityBitmask.HAVE_TRUST_CENTER_EUI64
        not in security.bitmask
    )


def test_zha_security_hashed_nonstandard_tclk_warning(network_info, caplog):
    # Nothing should be logged normally
    with caplog.at_level(logging.WARNING):
        util.zha_security(
            network_info=network_info,
            use_hashed_tclk=True,
        )

    assert "Only the well-known TC Link Key is supported" not in caplog.text

    # But it will be when a non-standard TCLK is used along with TCLK hashing
    with caplog.at_level(logging.WARNING):
        util.zha_security(
            network_info=network_info.replace(
                tc_link_key=network_info.tc_link_key.replace(
                    key=t.KeyData(b"ANonstandardTCLK")
                )
            ),
            use_hashed_tclk=True,
        )

    assert "Only the well-known TC Link Key is supported" in caplog.text


def test_ezsp_key_to_zigpy_key(zigpy_key, ezsp_key, ezsp_mock):
    assert util.ezsp_key_to_zigpy_key(ezsp_key, ezsp_mock) == zigpy_key


def test_zigpy_key_to_ezsp_key(zigpy_key, ezsp_key, ezsp_mock):
    assert util.zigpy_key_to_ezsp_key(zigpy_key, ezsp_mock) == ezsp_key


def test_map_rssi_to_energy():
    assert 0 <= util.map_rssi_to_energy(-200) <= 0.01
    assert 254 <= util.map_rssi_to_energy(100) <= 255

    # Make sure the two functions are inverses
    for rssi in range(-100, 100):
        energy = util.map_rssi_to_energy(rssi)
        assert abs(util.map_energy_to_rssi(energy) - rssi) < 0.1



from . import types as t

COMMANDS = {
    # 4. Configuration frames
    "version": (0x0000, (t.uint8_t,), (t.uint8_t, t.uint8_t, t.uint16_t)),
    "getConfigurationValue": (0x0052, (t.EzspConfigId,), (t.EzspStatus, t.uint16_t)),
    "setConfigurationValue": (0x0053, (t.EzspConfigId, t.uint16_t), (t.EzspStatus,)),
    "addEndpoint": (
        0x0002,
        (
            t.uint8_t,
            t.uint16_t,
            t.uint16_t,
            t.uint8_t,
            t.uint8_t,
            t.uint8_t,
            t.List[t.uint16_t],
            t.List[t.uint16_t],
        ),
        (t.EzspStatus,),
    ),
    "setPolicy": (0x0055, (t.EzspPolicyId, t.uint16_t), (t.EzspStatus,)),
    "getPolicy": (0x0056, (t.EzspPolicyId,), (t.EzspStatus, t.EzspDecisionId)),
    "sendPanIdUpdate": (0x0057, (t.EmberPanId,), (t.Bool,)),
    "getValue": (0x00AA, (t.EzspValueId,), (t.EzspStatus, t.LVBytes)),
    "getExtendedValue": (
        0x0003,
        (t.EzspExtendedValueId, t.uint32_t),
        (t.EzspStatus, t.LVBytes),
    ),
    "setValue": (0x00AB, (t.EzspValueId, t.LVBytes), (t.EzspStatus,)),
    "setGpioCurrentConfiguration": (
        0x00AC,
        (t.uint8_t, t.uint8_t, t.uint8_t),
        (t.EzspStatus,),
    ),
    "setGpioPowerUpDownConfiguration": (
        0x00AD,
        (t.uint8_t, t.uint8_t, t.uint8_t, t.uint8_t, t.uint8_t),
        (t.EzspStatus,),
    ),
    # 5. Utilities Frames
    "nop": (0x0005, (), ()),
    "echo": (0x0081, (t.LVBytes,), (t.LVBytes,)),
    "invalidCommand": (0x0058, (), (t.EzspStatus,)),
    "callback": (0x0006, (), ()),
    "noCallbacks": (0x0007, (), ()),
    "setToken": (0x0009, (t.uint8_t, t.FixedList[t.uint8_t, 8]), (t.EmberStatus,)),
    "getToken": (0x000A, (t.uint8_t,), (t.EmberStatus, t.FixedList[t.uint8_t, 8])),
    "getMfgToken": (0x000B, (t.EzspMfgTokenId,), (t.LVBytes,)),
    "setMfgToken": (0x000C, (t.EzspMfgTokenId, t.LVBytes), (t.EmberStatus,)),
    "stackTokenChangedHandler": (0x000D, (), (t.uint16_t,)),
    "getRandomNumber": (0x0049, (), (t.EmberStatus, t.uint16_t)),
    "setTimer": (
        0x000E,
        (t.uint8_t, t.uint16_t, t.EmberEventUnits, t.Bool),
        (t.EmberStatus,),
    ),
    "getTimer": (0x004E, (t.uint8_t,), (t.uint16_t, t.EmberEventUnits, t.Bool)),
    "timerHandler": (0x000F, (), (t.uint8_t,)),
    "debugWrite": (0x0012, (t.Bool, t.LVBytes), (t.EmberStatus,)),
    "readAndClearCounters": (
        0x0065,
        (),
        (t.FixedList[t.uint16_t, len(t.EmberCounterType)],),
    ),
    "readCounters": (
        0x00F1,
        (),
        (t.FixedList[t.uint16_t, len(t.EmberCounterType)],),
    ),
    "counterRolloverHandler": (0x00F2, (), (t.EmberCounterType,)),
    "delayTest": (0x009D, (t.uint16_t,), ()),
    "getLibraryStatus": (0x0001, (t.EmberLibraryId,), (t.EmberLibraryStatus,)),
    "getXncpInfo": (0x0013, (), (t.EmberStatus, t.uint16_t, t.uint16_t)),
    "customFrame": (0x0047, (t.LVBytes,), (t.EmberStatus, t.LVBytes)),
    "customFrameHandler": (0x0054, (), (t.LVBytes,)),
    "getEui64": (0x0026, (), (t.EUI64,)),
    "getNodeId": (0x0027, (), (t.EmberNodeId,)),
    "getPhyInterfaceCount": (0x00FC, (), (t.uint8_t,)),
    "getTrueRandomEntropySource": (0x004F, (), (t.EmberEntropySource,)),
    "networkInit": (0x0017, (t.EmberNetworkInitStruct,), (t.EmberStatus,)),
    # 6. Networking Frames
    "setManufacturerCode": (0x0015, (t.uint16_t,), ()),
    "setPowerDescriptor": (0x0016, (t.uint16_t,), ()),
    "networkState": (0x0018, (), (t.EmberNetworkStatus,)),
    "stackStatusHandler": (0x0019, (), (t.EmberStatus,)),
    "startScan": (
        0x001A,
        (t.EzspNetworkScanType, t.uint32_t, t.uint8_t),
        (t.EmberStatus,),
    ),
    "energyScanResultHandler": (0x0048, (), (t.uint8_t, t.int8s)),
    "networkFoundHandler": (0x001B, (), (t.EmberZigbeeNetwork, t.uint8_t, t.int8s)),
    "scanCompleteHandler": (0x001C, (), (t.uint8_t, t.EmberStatus)),
    "unusedPanIdFoundHandler": (0x00D2, (), (t.uint8_t, t.EmberStatus)),
    "findUnusedPanId": (0x00D3, (t.Channels, t.uint8_t), (t.EmberStatus,)),
    "stopScan": (0x001D, (), (t.EmberStatus,)),
    "formNetwork": (0x001E, (t.EmberNetworkParameters,), (t.EmberStatus,)),
    "joinNetwork": (
        0x001F,
        (t.EmberNodeType, t.EmberNetworkParameters),
        (t.EmberStatus,),
    ),
    "joinNetworkDirectly": (
        0x003B,
        (t.EmberNodeType, t.EmberBeaconData, t.int8s, t.Bool),
        (t.EmberStatus,),
    ),
    "leaveNetwork": (0x0020, (), (t.EmberStatus,)),
    "findAndRejoinNetwork": (0x0021, (t.Bool, t.uint32_t), (t.EmberStatus,)),
    "permitJoining": (0x0022, (t.uint8_t,), (t.EmberStatus,)),
    "childJoinHandler": (
        0x0023,
        (),
        (t.uint8_t, t.Bool, t.EmberNodeId, t.EUI64, t.EmberNodeType),
    ),
    "energyScanRequest": (
        0x009C,
        (t.EmberNodeId, t.Channels, t.uint8_t, t.uint16_t),
        (t.EmberStatus,),
    ),
    "getNetworkParameters": (
        0x0028,
        (),
        (t.EmberStatus, t.EmberNodeType, t.EmberNetworkParameters),
    ),
    "getRadioParameters": (
        0x00FD,
        (t.uint8_t,),
        (t.EmberStatus, t.EmberNodeType, t.EmberNetworkParameters),
    ),
    "getParentChildParameters": (0x0029, (), (t.uint8_t, t.EUI64, t.EmberNodeId)),
    "getChildData": (
        0x004A,
        (t.uint8_t,),
        (t.EmberStatus, t.EmberChildData),
    ),
    "getSourceRouteTableTotalSize": (0x00C3, (), (t.uint8_t,)),
    "getSourceRouteTableFilledSize": (0x00C2, (), (t.uint8_t,)),
    "getSourceRouteTableEntry": (
        0x00C1,
        (t.uint8_t,),
        (t.EmberStatus, t.EmberNodeId, t.uint8_t),
    ),
    "getNeighbor": (0x0079, (t.uint8_t,), (t.EmberStatus, t.EmberNeighborTableEntry)),
    "getNeighborFrameCounter": (0x003E, (t.EUI64,), (t.EmberStatus, t.uint32_t)),
    "setRoutingShortcutThreshold": (0x00D0, (t.uint8_t,), (t.EmberStatus,)),
    "getRoutingShortcutThreshold": (0x00D1, (), (t.uint8_t,)),
    "neighborCount": (0x007A, (), (t.uint8_t,)),
    "getRouteTableEntry": (
        0x007B,
        (t.uint8_t,),
        (t.EmberStatus, t.EmberRouteTableEntry),
    ),
    "setRadioPower": (0x0099, (t.int8s,), (t.EmberStatus,)),
    "setRadioChannel": (0x009A, (t.uint8_t,), (t.EmberStatus,)),
    "setConcentrator": (
        0x0010,
        (t.Bool, t.uint16_t, t.uint16_t, t.uint16_t, t.uint8_t, t.uint8_t, t.uint8_t),
        (t.EmberStatus,),
    ),
    "setBrokenRouteErrorCode": (0x0011, (t.uint8_t,), (t.EmberStatus,)),
    "multiPhyStart": (
        0x00F8,
        (t.uint8_t, t.uint8_t, t.uint8_t, t.int8s, t.EmberMultiPhyNwkConfig),
        (t.EmberStatus,),
    ),
    "multiPhyStop": (0x00F9, (t.uint8_t,), (t.EmberStatus,)),
    "multiPhySetRadioPower": (0x00FA, (t.uint8_t, t.int8s), (t.EmberStatus,)),
    "sendLinkPowerDeltaRequest": (0x00F7, (), (t.EmberStatus,)),
    "multiPhySetRadioChannel": (
        0x00FB,
        (t.uint8_t, t.uint8_t, t.uint8_t),
        (t.EmberStatus,),
    ),
    "getDutyCycleState": (0x0035, (), (t.EmberStatus, t.EmberDutyCycleState)),
    "setDutyCycleLimitsInStack": (
        0x0040,
        (t.EmberDutyCycleLimits,),
        (t.EmberStatus, t.EmberDutyCycleState),
    ),
    "getDutyCycleLimits": (0x004B, (), (t.EmberStatus, t.EmberDutyCycleLimits)),
    "getCurrentDutyCycle": (
        0x004C,
        (t.uint8_t,),
        (t.EmberStatus, t.FixedList[t.uint8_t, 134]),
    ),
    "dutyCycleHandler": (
        0x004D,
        (),
        (
            t.uint8_t,
            t.uint8_t,
            t.EmberDutyCycleState,
            t.uint8_t,
            t.EmberPerDeviceDutyCycle,
        ),
    ),
    "getFirstBeacon": (0x003D, (), (t.EmberStatus, t.EmberBeaconIterator)),
    "getNextBeacon": (0x0004, (), (t.EmberStatus, t.EmberBeaconData)),
    "getNumStoredBeacons": (0x0008, (), (t.uint8_t,)),
    "clearStoredBeacons": (0x003C, (), ()),
    "setLogicalAndRadioChannel": (0x00B9, (t.uint8_t,), (t.EmberStatus,)),
    "getLogicalChannel": (0x00BA, (), (t.uint8_t,)),
    # 7. Binding Frames
    "clearBindingTable": (0x002A, (), (t.EmberStatus,)),
    "setBinding": (0x002B, (t.uint8_t, t.EmberBindingTableEntry), (t.EmberStatus,)),
    "getBinding": (0x002C, (t.uint8_t,), (t.EmberStatus, t.EmberBindingTableEntry)),
    "deleteBinding": (0x002D, (t.uint8_t,), (t.EmberStatus,)),
    "bindingIsActive": (0x002E, (t.uint8_t,), (t.Bool,)),
    "getBindingRemoteNodeId": (0x002F, (t.uint8_t,), (t.EmberNodeId,)),
    "setBindingRemoteNodeId": (0x0030, (t.uint8_t, t.EmberNodeId), ()),
    "remoteSetBindingHandler": (
        0x0031,
        (),
        (t.EmberBindingTableEntry, t.uint8_t, t.EmberStatus),
    ),
    "remoteDeleteBindingHandler": (0x0032, (), (t.uint8_t, t.EmberStatus)),
    # 8. Messaging Frames
    "maximumPayloadLength": (0x0033, (), (t.uint8_t,)),
    "sendUnicast": (
        0x0034,
        (
            t.EmberOutgoingMessageType,
            t.EmberNodeId,
            t.EmberApsFrame,
            t.uint8_t,
            t.LVBytes,
        ),
        (t.EmberStatus, t.uint8_t),
    ),
    "sendBroadcast": (
        0x0036,
        (t.EmberNodeId, t.EmberApsFrame, t.uint8_t, t.uint8_t, t.LVBytes),
        (t.EmberStatus, t.uint8_t),
    ),
    "proxyBroadcast": (
        0x0037,
        (
            t.EmberNodeId,
            t.EmberNodeId,
            t.uint8_t,
            t.EmberApsFrame,
            t.uint8_t,
            t.uint8_t,
            t.LVBytes,
        ),
        (t.EmberStatus, t.uint8_t),
    ),
    "sendMulticast": (
        0x0038,
        (t.EmberApsFrame, t.uint8_t, t.uint8_t, t.uint8_t, t.LVBytes),
        (t.EmberStatus, t.uint8_t),
    ),
    "sendMulticastWithAlias": (
        0x3A,
        (
            t.EmberApsFrame,
            t.uint8_t,
            t.uint8_t,
            t.uint16_t,
            t.uint8_t,
            t.uint8_t,
            t.LVBytes,
        ),
        (t.EmberStatus, t.uint8_t),
    ),
    "sendReply": (
        0x0039,
        (t.EmberNodeId, t.EmberApsFrame, t.LVBytes),
        (t.EmberStatus,),
    ),
    "messageSentHandler": (
        0x003F,
        (),
        (
            t.EmberOutgoingMessageType,
            t.uint16_t,
            t.EmberApsFrame,
            t.uint8_t,
            t.EmberStatus,
            t.LVBytes,
        ),
    ),
    "sendManyToOneRouteRequest": (0x0041, (t.uint16_t, t.uint8_t), (t.EmberStatus,)),
    "pollForData": (
        0x0042,
        (t.uint16_t, t.EmberEventUnits, t.uint8_t),
        (t.EmberStatus,),
    ),
    "pollCompleteHandler": (0x0043, (), (t.EmberStatus,)),
    "pollHandler": (0x0044, (), (t.EmberNodeId,)),
    "incomingSenderEui64Handler": (0x0062, (), (t.EUI64,)),
    "incomingMessageHandler": (
        0x0045,
        (),
        (
            t.EmberIncomingMessageType,
            t.EmberApsFrame,
            t.uint8_t,
            t.int8s,
            t.EmberNodeId,
            t.uint8_t,
            t.uint8_t,
            t.LVBytes,
        ),
    ),
    "incomingRouteRecordHandler": (
        0x0059,
        (),
        (t.EmberNodeId, t.EUI64, t.uint8_t, t.int8s, t.LVList[t.EmberNodeId]),
    ),
    "incomingNetworkStatusHandler": (
        0x00C4,
        (),
        tuple({"errorCode": t.EmberStackError, "target": t.EmberNodeId}.values()),
    ),
    "setSourceRoute": (
        0x00AE,
        (t.EmberNodeId, t.LVList[t.EmberNodeId]),
        (t.EmberStatus,),
    ),
    "setSourceRouteDiscoveryMode": (0x005A, (t.uint8_t,), (t.uint32_t,)),
    "incomingManyToOneRouteRequestHandler": (
        0x007D,
        (),
        (t.EmberNodeId, t.EUI64, t.uint8_t),
    ),
    "incomingRouteErrorHandler": (0x0080, (), (t.EmberStatus, t.EmberNodeId)),
    "unicastCurrentNetworkKey": (
        0x0050,
        (t.EmberNodeId, t.EUI64, t.EmberNodeId),
        (t.EmberStatus,),
    ),
    "addressTableEntryIsActive": (0x005B, (t.uint8_t,), (t.Bool,)),
    "setAddressTableRemoteEui64": (0x005C, (t.uint8_t, t.EUI64), (t.EmberStatus,)),
    "setAddressTableRemoteNodeId": (0x005D, (t.uint8_t, t.EmberNodeId), ()),
    "getAddressTableRemoteEui64": (0x005E, (t.uint8_t,), (t.EUI64,)),
    "getAddressTableRemoteNodeId": (0x005F, (t.uint8_t,), (t.EmberNodeId,)),
    "setExtendedTimeout": (0x007E, (t.EUI64, t.Bool), ()),
    "getExtendedTimeout": (0x007F, (t.EUI64,), (t.Bool,)),
    "replaceAddressTableEntry": (
        0x0082,
        (t.uint8_t, t.EUI64, t.EmberNodeId, t.Bool),
        (t.EmberStatus, t.EUI64, t.EmberNodeId, t.Bool),
    ),
    "lookupNodeIdByEui64": (0x0060, (t.EUI64,), (t.EmberNodeId,)),
    "lookupEui64ByNodeId": (0x0061, (t.EmberNodeId,), (t.EmberStatus, t.EUI64)),
    "getMulticastTableEntry": (
        0x0063,
        (t.uint8_t,),
        (t.EmberStatus, t.EmberMulticastTableEntry),
    ),
    "setMulticastTableEntry": (
        0x0064,
        (t.uint8_t, t.EmberMulticastTableEntry),
        (t.EmberStatus,),
    ),
    "idConflictHandler": (0x007C, (), (t.EmberNodeId,)),
    "writeNodeData": (0x00FE, (t.Bool,), (t.EmberStatus,)),
    "sendRawMessage": (0x0096, (t.LVBytes,), (t.EmberStatus,)),
    "sendRawMessageExtended": (
        0x0051,
        (t.LVBytes, t.uint8_t, t.Bool),
        (t.EmberStatus,),
    ),
    "macPassthroughMessageHandler": (
        0x0097,
        (),
        (t.EmberMacPassthroughType, t.uint8_t, t.int8s, t.LVBytes),
    ),
    "macFilterMatchMessageHandler": (
        0x0046,
        (),
        (t.uint8_t, t.EmberMacPassthroughType, t.uint8_t, t.int8s, t.LVBytes),
    ),
    "rawTransmitCompleteHandler": (0x0098, (), (t.EmberStatus,)),
    "setMacPollFailureWaitTime": (0x00F4, (t.uint8_t,), ()),
    "setBeaconClassificationParams": (
        0x00EF,
        (),
        (t.EmberStatus, t.EmberBeaconClassificationParams),
    ),
    "getBeaconClassificationParams": (
        0x00F3,
        (),
        (t.EmberStatus, t.EmberBeaconClassificationParams),
    ),
    # 9. Security Frames
    "setInitialSecurityState": (
        0x0068,
        (t.EmberInitialSecurityState,),
        (t.EmberStatus,),
    ),
    "getCurrentSecurityState": (
        0x0069,
        (),
        (t.EmberStatus, t.EmberCurrentSecurityState),
    ),
    "getKey": (0x006A, (t.EmberKeyType,), (t.EmberStatus, t.EmberKeyStruct)),
    "switchNetworkKeyHandler": (0x006E, (), (t.uint8_t,)),
    "getKeyTableEntry": (0x0071, (t.uint8_t,), (t.EmberStatus, t.EmberKeyStruct)),
    "setKeyTableEntry": (
        0x0072,
        (t.uint8_t, t.EUI64, t.Bool, t.KeyData),
        (t.EmberStatus,),
    ),
    "findKeyTableEntry": (0x0075, (t.EUI64, t.Bool), (t.uint8_t,)),
    "addOrUpdateKeyTableEntry": (
        0x0066,
        (t.EUI64, t.Bool, t.KeyData),
        (t.EmberStatus,),
    ),
    "sendTrustCenterLinkKey": (0x0067, (t.EmberNodeId, t.EUI64), (t.EmberStatus,)),
    "eraseKeyTableEntry": (0x0076, (t.uint8_t,), (t.EmberStatus,)),
    "clearKeyTable": (0x00B1, (), (t.EmberStatus,)),
    "requestLinkKey": (0x0014, (t.EUI64,), (t.EmberStatus,)),
    "updateTcLinkKey": (0x006C, (t.uint8_t,), (t.EmberStatus,)),
    "zigbeeKeyEstablishmentHandler": (0x009B, (), (t.EUI64, t.EmberKeyStatus)),
    "addTransientLinkKey": (0x00AF, (t.EUI64, t.KeyData), (t.EmberStatus,)),
    "clearTransientLinkKeys": (0x006B, (), ()),
    "getTransientLinkKey": (
        0x00CE,
        (t.EUI64,),
        (t.EmberStatus, t.EmberTransientKeyData),
    ),
    "getTransientKeyTableEntry": (
        0x006D,
        (t.uint8_t,),
        (t.EmberStatus, t.EmberTransientKeyData),
    ),
    # 10. Trust Center Frames
    "trustCenterJoinHandler": (
        0x0024,
        (),
        (
            t.EmberNodeId,
            t.EUI64,
            t.EmberDeviceUpdate,
            t.EmberJoinDecision,
            t.EmberNodeId,
        ),
    ),
    "broadcastNextNetworkKey": (0x0073, (t.KeyData,), (t.EmberStatus,)),
    "broadcastNetworkKeySwitch": (0x0074, (), (t.EmberStatus,)),
    "becomeTrustCenter": (0x0077, (t.KeyData,), (t.EmberStatus,)),
    "aesMmoHash": (
        0x006F,
        (t.EmberAesMmoHashContext, t.Bool, t.LVBytes),
        (t.EmberStatus, t.EmberAesMmoHashContext),
    ),
    "removeDevice": (
        0x00A8,
        (t.EmberNodeId, t.EUI64, t.EUI64),
        (t.EmberStatus,),
    ),
    "unicastNwkKeyUpdate": (
        0x00A9,
        (t.EmberNodeId, t.EUI64, t.KeyData),
        (t.EmberStatus,),
    ),
    # 11. Certificate Based Key Exchange (CBKE) Frames
    "generateCbkeKeys": (0x00A4, (), (t.EmberStatus,)),
    "generateCbkeKeysHandler": (0x009E, (), (t.EmberStatus, t.EmberPublicKeyData)),
    "calculateSmacs": (
        0x009F,
        (t.Bool, t.EmberCertificateData, t.EmberPublicKeyData),
        (t.EmberStatus,),
    ),
    "calculateSmacsHandler": (
        0x00A0,
        (),
        (t.EmberStatus, t.EmberSmacData, t.EmberSmacData),
    ),
    "generateCbkeKeys283k1": (0x00E8, (), (t.EmberStatus,)),
    "generateCbkeKeysHandler283k1": (
        0x00E9,
        (),
        (t.EmberStatus, t.EmberPublicKey283k1Data),
    ),
    "calculateSmacs283k1": (
        0x00EA,
        (t.Bool, t.EmberCertificate283k1Data, t.EmberPublicKey283k1Data),
        (t.EmberStatus,),
    ),
    "calculateSmacsHandler283k1": (
        0x00EB,
        (),
        (t.EmberStatus, t.EmberSmacData, t.EmberSmacData),
    ),
    "clearTemporaryDataMaybeStoreLinkKey": (0x00A1, (t.Bool,), (t.EmberStatus,)),
    "clearTemporaryDataMaybeStoreLinkKey283k1": (0x00EE, (t.Bool,), (t.EmberStatus,)),
    "getCertificate": (0x00A5, (), (t.EmberStatus, t.EmberCertificateData)),
    "getCertificate283k1": (0x00EC, (), (t.EmberStatus, t.EmberCertificate283k1Data)),
    "dsaSign": (0x00A6, (t.LVBytes,), (t.EmberStatus,)),  # Deprecated
    "dsaSignHandler": (0x00A7, (), (t.EmberStatus, t.LVBytes)),  # Deprecated
    "dsaVerify": (
        0x00A3,
        (t.EmberMessageDigest, t.EmberCertificateData, t.EmberSignatureData),
        (t.EmberStatus,),
    ),
    "dsaVerifyHandler": (0x0078, (), (t.EmberStatus,)),
    "dsaVerify283k1": (
        0x00B0,
        (t.EmberMessageDigest, t.EmberCertificate283k1Data, t.EmberSignature283k1Data),
        (t.EmberStatus,),
    ),
    "setPreinstalledCbkeData": (
        0x00A2,
        (t.EmberPublicKeyData, t.EmberCertificateData, t.EmberPrivateKeyData),
        (t.EmberStatus,),
    ),
    "setPreinstalledCbkeData283k1": (
        0x00ED,
        (
            t.EmberPublicKey283k1Data,
            t.EmberCertificate283k1Data,
            t.EmberPrivateKey283k1Data,
        ),
        (t.EmberStatus,),
    ),
    # 12. Mfglib Frames
    "mfglibStart": (0x0083, (t.Bool,), (t.EmberStatus,)),
    "mfglibEnd": (0x0084, (), (t.EmberStatus,)),
    "mfglibStartTone": (0x0085, (), (t.EmberStatus,)),
    "mfglibStopTone": (0x0086, (), (t.EmberStatus,)),
    "mfglibStartStream": (0x0087, (), (t.EmberStatus,)),
    "mfglibStopStream": (0x0088, (), (t.EmberStatus,)),
    "mfglibSendPacket": (0x0089, (t.LVBytes,), (t.EmberStatus,)),
    "mfglibSetChannel": (0x008A, (t.uint8_t,), (t.EmberStatus,)),
    "mfglibGetChannel": (0x008B, (), (t.uint8_t,)),
    "mfglibSetPower": (0x008C, (t.uint16_t, t.int8s), (t.EmberStatus,)),
    "mfglibGetPower": (0x008D, (), (t.int8s,)),
    "mfglibRxHandler": (0x008E, (), (t.uint8_t, t.int8s, t.LVBytes)),
    # 13. Bootloader Frames
    "launchStandaloneBootloader": (0x008F, (t.uint8_t,), (t.EmberStatus,)),
    "sendBootloadMessage": (
        0x0090,
        (t.Bool, t.EUI64, t.LVBytes),
        (t.EmberStatus,),
    ),
    "getStandaloneBootloaderVersionPlatMicroPhy": (
        0x0091,
        (),
        (t.uint16_t, t.uint8_t, t.uint8_t, t.uint8_t),
    ),
    "incomingBootloadMessageHandler": (
        0x0092,
        (),
        (t.EUI64, t.uint8_t, t.int8s, t.LVBytes),
    ),
    "bootloadTransmitCompleteHandler": (0x0093, (), (t.EmberStatus, t.LVBytes)),
    "aesEncrypt": (
        0x0094,
        (t.FixedList[t.uint8_t, 16], t.FixedList[t.uint8_t, 16]),
        (t.FixedList[t.uint8_t, 16],),
    ),
    "overrideCurrentChannel": (0x0095, (t.uint8_t,), (t.EmberStatus,)),
    # 14. ZLL Frames
    "zllNetworkOps": (
        0x00B2,
        (t.EmberZllNetwork, t.EzspZllNetworkOperation, t.int8s),
        (t.EmberStatus,),
    ),
    "zllSetInitialSecurityState": (
        0x00B3,
        (t.KeyData, t.EmberZllInitialSecurityState),
        (t.EmberStatus,),
    ),
    "zllSetSecurityStateWithoutKey": (
        0x00CF,
        (t.EmberZllInitialSecurityState,),
        (t.EmberStatus,),
    ),
    "zllStartScan": (0x00B4, (t.Channels, t.int8s, t.EmberNodeType), (t.EmberStatus,)),
    "zllSetRxOnWhenIdle": (0x00B5, (t.uint32_t,), (t.EmberStatus,)),
    "zllNetworkFoundHandler": (
        0x00B6,
        (),
        (t.EmberZllNetwork, t.Bool, t.EmberZllDeviceInfoRecord, t.uint8_t, t.int8s),
    ),
    "zllScanCompleteHandler": (0x00B7, (), (t.EmberStatus,)),
    "zllAddressAssignmentHandler": (
        0x00B8,
        (),
        (t.EmberZllAddressAssignment, t.uint8_t, t.int8s),
    ),
    "zllTouchLinkTargetHandler": (0x00BB, (), (t.EmberZllNetwork,)),
    "zllGetTokens": (
        0x00BC,
        (),
        (t.EmberTokTypeStackZllData, t.EmberTokTypeStackZllSecurity),
    ),
    "zllSetDataToken": (0x00BD, (t.EmberTokTypeStackZllData,), ()),
    "zllSetNonZllNetwork": (0x00BF, (), ()),
    "isZllNetwork": (0x00BE, (), (t.Bool,)),
    "zllSetRadioIdleMode": (0x00D4, (t.EmberRadioPowerMode,), ()),
    "setZllNodeType": (0x00D5, (t.EmberNodeType,), ()),
    "setZllAdditionalState": (0x00D6, (t.uint16_t,), ()),
    "zllOperationInProgress": (0x00D7, (), (t.Bool,)),
    "zllRxOnWhenIdleGetActive": (0x00D8, (), (t.Bool,)),
    "getZllPrimaryChannelMask": (0x00D9, (), (t.Channels,)),
    "getZllSecondaryChannelMask": (0x00DA, (), (t.Channels,)),
    "setZllPrimaryChannelMask": (0x00DB, (t.Channels,), ()),
    "setZllSecondaryChannelMask": (0x00DC, (t.Channels,), ()),
    "zllClearTokens": (0x0025, (), ()),
    # 15 WWAH Frames
    "setParentClassificationEnabled": (0x00E7, (t.Bool,), ()),
    "getParentClassificationEnabled": (0x00F0, (), (t.Bool,)),
    "setLongUpTime": (0x00E3, (t.Bool,), ()),
    "setHubConnectivity": (0x00E4, (t.Bool,), ()),
    "isUpTimeLong": (0x00E5, (), (t.Bool,)),
    "isHubConnected": (0x00E6, (), (t.Bool,)),
    # 16 Green Power Frames
    "gpProxyTableProcessGpPairing": (
        0x00C9,
        (
            t.uint32_t,
            t.EmberGpAddress,
            t.uint8_t,
            t.uint16_t,
            t.uint16_t,
            t.uint16_t,
            t.EUI64,
            t.KeyData,
        ),
        (),
    ),
    "dGpSend": (
        0x00C6,
        (t.Bool, t.Bool, t.EmberGpAddress, t.uint8_t, t.LVBytes, t.uint8_t, t.uint16_t),
        (t.EmberStatus,),
    ),
    "dGpSentHandler": (0x00C7, (), (t.EmberStatus, t.uint8_t)),
    "gpepIncomingMessageHandler": (
        0x00C5,
        (),
        (
            t.EmberStatus,  # status
            t.uint8_t,  # gpd link
            t.uint8_t,  # sequence number
            t.uint8_t,  # addrType
            t.uint32_t,  # address
            t.uint32_t,  # src ID
            t.uint8_t,  # endpoint
            t.EmberGpSecurityLevel,
            t.EmberGpKeyType,
            t.Bool,  # auto commissioning
            t.Bool,  # rx capable?
            t.uint32_t,  # security frame counter
            t.uint8_t,  # gpd command id
            t.uint32_t,  # MIC
            t.uint8_t,  # proxy table index
            t.LVBytes,  # optional payload
        ),
    ),
    "gpProxyTableGetEntry": (
        0x00C8,
        (t.uint8_t,),
        (t.EmberStatus, t.EmberGpProxyTableEntry),
    ),
    "gpProxyTableLookup": (0x00C0, (t.EmberGpAddress,), (t.uint8_t,)),
    "gpSinkTableGetEntry": (
        0x00DD,
        (t.uint8_t,),
        (t.EmberStatus, t.EmberGpSinkTableEntry),
    ),
    "gpSinkTableLookup": (0x00DE, (t.EmberGpAddress,), (t.uint8_t,)),
    "gpSinkTableSetEntry": (
        0x00DF,
        (t.uint8_t, t.EmberGpSinkTableEntry),
        (t.EmberStatus,),
    ),
    "gpSinkTableRemoveEntry": (0x00E0, (t.uint8_t,), ()),
    "gpSinkTableFindOrAllocateEntry": (0x00E1, (t.EmberGpAddress,), (t.uint8_t,)),
    "gpSinkTableClearAll": (0x00E2, (), ()),
    "gpSinkTableInit": (0x0070, (), ()),
    # 16 Secure EZSP Frames
    "setSecurityKey": (
        0x00CA,
        (t.KeyData, t.SecureEzspSecurityType),
        (t.EzspStatus,),
    ),
    "setSecurityParameters": (
        0x00CB,
        (t.SecureEzspSecurityLevel, t.SecureEzspRandomNumber),
        (t.EzspStatus, t.SecureEzspRandomNumber),
    ),
    "resetToFactoryDefaults": (0x00CC, (), (t.EzspStatus,)),
    "getSecurityKeyStatus": (0x00CD, (), (t.EzspStatus, t.SecureEzspSecurityType)),
}



import voluptuous as vol

from bellows.config import cv_optional_int, cv_uint16, extend_vol_schema

from ..v4.config import EZSP_POLICIES_SHARED
from ..v5 import config as v5_config
from .types import EzspConfigId, EzspPolicyId

EZSP_SCHEMA = extend_vol_schema(
    v5_config.EZSP_SCHEMA,
    {
        vol.Optional(EzspConfigId.CONFIG_RETRY_QUEUE_SIZE.name): cv_optional_int(min=0),
        vol.Optional(
            EzspConfigId.CONFIG_NEW_BROADCAST_ENTRY_THRESHOLD.name
        ): cv_optional_int(min=0),
        vol.Optional(
            EzspConfigId.CONFIG_BROADCAST_MIN_ACKS_NEEDED.name
        ): cv_optional_int(min=0),
        vol.Optional(
            EzspConfigId.CONFIG_TC_REJOINS_USING_WELL_KNOWN_KEY_TIMEOUT_S.name
        ): cv_optional_int(min=0),
    },
)

del EZSP_SCHEMA[v5_config.EzspConfigId.CONFIG_MOBILE_NODE_POLL_TIMEOUT.name]
del EZSP_SCHEMA[v5_config.EzspConfigId.CONFIG_RESERVED_MOBILE_CHILD_ENTRIES.name]
del EZSP_SCHEMA[v5_config.EzspConfigId.CONFIG_RF4CE_PAIRING_TABLE_SIZE.name]
del EZSP_SCHEMA[
    v5_config.EzspConfigId.CONFIG_RF4CE_PENDING_OUTGOING_PACKET_TABLE_SIZE.name
]

EZSP_POLICIES_SCH = {
    **EZSP_POLICIES_SHARED,
    **{vol.Optional(policy.name): cv_uint16 for policy in EzspPolicyId},
}



import voluptuous as vol

from bellows.config import cv_optional_int, cv_uint16, extend_vol_schema

from ..v4.config import EZSP_POLICIES_SHARED
from ..v6 import config as v6_config
from .types import EzspConfigId, EzspPolicyId

EZSP_SCHEMA = extend_vol_schema(
    v6_config.EZSP_SCHEMA,
    {
        vol.Optional(
            EzspConfigId.CONFIG_END_DEVICE_POLL_TIMEOUT.name, default=8
        ): cv_optional_int(min=0, max=14),
        vol.Optional(
            EzspConfigId.CONFIG_KEY_TABLE_SIZE.name, default=12
        ): cv_optional_int(min=0),
    },
)

del EZSP_SCHEMA["CONFIG_END_DEVICE_POLL_TIMEOUT_SHIFT"]

EZSP_POLICIES_SCH = {
    **EZSP_POLICIES_SHARED,
    **{vol.Optional(policy.name): cv_uint16 for policy in EzspPolicyId},
}



from __future__ import annotations

import abc
import asyncio
import binascii
from collections.abc import Coroutine
import dataclasses
import enum
import logging
import sys
import time
import typing

if sys.version_info[:2] < (3, 11):
    from async_timeout import timeout as asyncio_timeout  # pragma: no cover
else:
    from asyncio import timeout as asyncio_timeout  # pragma: no cover

from zigpy.types import BaseDataclassMixin

import bellows.types as t

_LOGGER = logging.getLogger(__name__)

MAX_BUFFER_SIZE = 1024


class Reserved(enum.IntEnum):
    FLAG = 0x7E  # Marks end of frame
    ESCAPE = 0x7D
    XON = 0x11  # Resume transmission
    XOFF = 0x13  # Stop transmission
    SUBSTITUTE = 0x18  # Replaces a byte received with a low-level communication error
    CANCEL = 0x1A  # Terminates a frame in progress


RESERVED_BYTES = frozenset(Reserved)
RESERVED_WITHOUT_ESCAPE = frozenset([v for v in Reserved if v != Reserved.ESCAPE])

# Initial value of t_rx_ack, the maximum time the NCP waits to receive acknowledgement
# of a DATA frame
T_RX_ACK_INIT = 1.6

# Minimum value of t_rx_ack
T_RX_ACK_MIN = 0.4

# Maximum value of t_rx_ack
T_RX_ACK_MAX = 3.2

# Delay before sending a non-piggybacked acknowledgement
T_TX_ACK_DELAY = 0.02

# Time from receiving an ACK or NAK with the nRdy flag set after which the NCP resumes
# sending callback frames to the host without requiring an ACK or NAK with the nRdy
# flag clear
T_REMOTE_NOTRDY = 1.0

# Maximum number of DATA frames the NCP can transmit without having received
# acknowledgements
TX_K = 1  # TODO: investigate why this cannot be raised without causing a firmware crash

# Maximum number of consecutive timeouts allowed while waiting to receive an ACK before
# going to the FAILED state. The value 0 prevents the NCP from entering the error state
# due to timeouts.
ACK_TIMEOUTS = 4


def generate_random_sequence(length: int) -> bytes:
    output = bytearray()
    rand = 0x42

    for _i in range(length):
        output.append(rand)

        if rand & 0b00000001 == 0:
            rand = rand >> 1
        else:
            rand = (rand >> 1) ^ 0xB8

    return output


# Since the sequence is static for every frame, we only need to generate it once
PSEUDO_RANDOM_DATA_SEQUENCE = generate_random_sequence(256)

if sys.version_info[:2] < (3, 12):
    create_eager_task = asyncio.create_task
else:
    _T = typing.TypeVar("T")

    def create_eager_task(
        coro: Coroutine[typing.Any, typing.Any, _T],
        *,
        name: str | None = None,
        loop: asyncio.AbstractEventLoop | None = None,
    ) -> asyncio.Task[_T]:
        """Create a task from a coroutine and schedule it to run immediately."""
        if loop is None:
            loop = asyncio.get_running_loop()

        return asyncio.Task(coro, loop=loop, name=name, eager_start=True)


class NcpState(enum.Enum):
    CONNECTED = "connected"
    FAILED = "failed"


class ParsingError(Exception):
    pass


class AshException(Exception):
    pass


class NotAcked(AshException):
    def __init__(self, frame: NakFrame) -> None:
        self.frame = frame

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}(frame={self.frame})>"


class NcpFailure(AshException):
    def __init__(self, code: t.NcpResetCode) -> None:
        self.code = code

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}(code={self.code})>"


class AshFrame(abc.ABC, BaseDataclassMixin):
    MASK: t.uint8_t
    MASK_VALUE: t.uint8_t

    @classmethod
    def from_bytes(cls, data: bytes) -> AshFrame:
        raise NotImplementedError()

    def to_bytes(self) -> bytes:
        raise NotImplementedError()

    @classmethod
    def _unwrap(cls, data: bytes) -> tuple[int, bytes]:
        if len(data) < 3:
            raise ParsingError(f"Frame is too short: {data!r}")

        computed_crc = binascii.crc_hqx(data[:-2], 0xFFFF).to_bytes(2, "big")

        if computed_crc != data[-2:]:
            raise ParsingError(
                f"Invalid CRC bytes in frame {data!r}:"
                f" expected {computed_crc.hex()}, got {data[-2:].hex()}"
            )

        return data[0], data[1:-2]

    @staticmethod
    def append_crc(data: bytes) -> bytes:
        return data + binascii.crc_hqx(data, 0xFFFF).to_bytes(2, "big")


@dataclasses.dataclass(frozen=True)
class DataFrame(AshFrame):
    MASK = 0b10000000
    MASK_VALUE = 0b00000000

    frm_num: int
    re_tx: bool
    ack_num: int
    ezsp_frame: bytes

    @staticmethod
    def _randomize(data: bytes) -> bytes:
        assert len(data) <= len(PSEUDO_RANDOM_DATA_SEQUENCE)
        return bytes([a ^ b for a, b in zip(data, PSEUDO_RANDOM_DATA_SEQUENCE)])

    @classmethod
    def from_bytes(cls, data: bytes) -> DataFrame:
        control, data = cls._unwrap(data)

        return cls(
            frm_num=(control & 0b01110000) >> 4,
            re_tx=(control & 0b00001000) >> 3,
            ack_num=(control & 0b00000111) >> 0,
            ezsp_frame=cls._randomize(data),
        )

    def to_bytes(self, *, randomize: bool = True) -> bytes:
        return self.append_crc(
            bytes(
                [
                    self.MASK_VALUE
                    | (self.frm_num) << 4
                    | (self.re_tx) << 3
                    | (self.ack_num) << 0
                ]
            )
            + self._randomize(self.ezsp_frame)
        )


@dataclasses.dataclass(frozen=True)
class AckFrame(AshFrame):
    MASK = 0b11100000
    MASK_VALUE = 0b10000000

    res: int
    ncp_ready: bool
    ack_num: int

    @classmethod
    def from_bytes(cls, data: bytes) -> AckFrame:
        control, data = cls._unwrap(data)

        return cls(
            res=(control & 0b00010000) >> 4,
            ncp_ready=(control & 0b00001000) >> 3,
            ack_num=(control & 0b00000111) >> 0,
        )

    def to_bytes(self) -> bytes:
        return self.append_crc(
            bytes(
                [
                    self.MASK_VALUE
                    | (self.res) << 4
                    | (self.ncp_ready) << 3
                    | (self.ack_num) << 0
                ]
            )
        )


@dataclasses.dataclass(frozen=True)
class NakFrame(AshFrame):
    MASK = 0b11100000
    MASK_VALUE = 0b10100000

    res: int
    ncp_ready: bool
    ack_num: int

    @classmethod
    def from_bytes(cls, data: bytes) -> AckFrame:
        control, data = cls._unwrap(data)

        return cls(
            res=(control & 0b00010000) >> 4,
            ncp_ready=(control & 0b00001000) >> 3,
            ack_num=(control & 0b00000111) >> 0,
        )

    def to_bytes(self) -> bytes:
        return self.append_crc(
            bytes(
                [
                    self.MASK_VALUE
                    | (self.res) << 4
                    | (self.ncp_ready) << 3
                    | (self.ack_num) << 0
                ]
            )
        )


@dataclasses.dataclass(frozen=True)
class RstFrame(AshFrame):
    MASK = 0b11111111
    MASK_VALUE = 0b11000000

    @classmethod
    def from_bytes(cls, data: bytes) -> RstFrame:
        control, data = cls._unwrap(data)

        if data:
            raise ParsingError(f"Invalid data for RST frame: {data!r}")

        return cls()

    def to_bytes(self) -> bytes:
        return self.append_crc(bytes([self.MASK_VALUE]))


@dataclasses.dataclass(frozen=True)
class RStackFrame(AshFrame):
    MASK = 0b11111111
    MASK_VALUE = 0b11000001

    version: t.uint8_t
    reset_code: t.NcpResetCode

    @classmethod
    def from_bytes(cls, data: bytes) -> RStackFrame:
        control, data = cls._unwrap(data)

        if len(data) != 2:
            raise ParsingError(f"Invalid data length for RSTACK frame: {data!r}")

        version = data[0]

        if version != 0x02:
            raise ParsingError(f"Invalid version for RSTACK frame: {data!r}")

        reset_code = t.NcpResetCode(data[1])

        return cls(
            version=version,
            reset_code=reset_code,
        )

    def to_bytes(self) -> bytes:
        return self.append_crc(bytes([self.MASK_VALUE, self.version, self.reset_code]))


@dataclasses.dataclass(frozen=True)
class ErrorFrame(AshFrame):
    MASK = 0b11111111
    MASK_VALUE = 0b11000010

    version: t.uint8_t
    reset_code: t.NcpResetCode

    # We do not want to inherit from `RStackFrame`
    from_bytes = classmethod(RStackFrame.from_bytes.__func__)
    to_bytes = RStackFrame.to_bytes


def parse_frame(
    data: bytes,
) -> DataFrame | AckFrame | NakFrame | RstFrame | RStackFrame | ErrorFrame:
    """Parse a frame from the given data, looking at the control byte."""
    control_byte = data[0]

    # In order of use
    for frame in [
        DataFrame,
        AckFrame,
        NakFrame,
        RstFrame,
        RStackFrame,
        ErrorFrame,
    ]:
        if control_byte & frame.MASK == frame.MASK_VALUE:
            return frame.from_bytes(data)
    else:
        raise ParsingError(f"Could not determine frame type: {data!r}")


class AshProtocol(asyncio.Protocol):
    def __init__(self, ezsp_protocol) -> None:
        self._ezsp_protocol = ezsp_protocol
        self._transport = None
        self._buffer = bytearray()
        self._discarding_until_next_flag: bool = False
        self._pending_data_frames: dict[int, asyncio.Future] = {}
        self._send_data_frame_semaphore = asyncio.Semaphore(TX_K)
        self._tx_seq: int = 0
        self._rx_seq: int = 0
        self._t_rx_ack = T_RX_ACK_INIT

        self._ncp_reset_code: t.NcpResetCode | None = None
        self._ncp_state: NcpState = NcpState.CONNECTED

    def connection_made(self, transport):
        self._transport = transport
        self._ezsp_protocol.connection_made(self)

    def connection_lost(self, exc):
        self._ezsp_protocol.connection_lost(exc)

    def eof_received(self):
        self._ezsp_protocol.eof_received()

    def close(self):
        if self._transport is not None:
            self._transport.close()

    @staticmethod
    def _stuff_bytes(data: bytes) -> bytes:
        """Stuff bytes for transmission"""
        out = bytearray()

        for c in data:
            if c in RESERVED_BYTES:
                out.extend([Reserved.ESCAPE, c ^ 0b00100000])
            else:
                out.append(c)

        return out

    @staticmethod
    def _unstuff_bytes(data: bytes) -> bytes:
        """Unstuff bytes after receipt"""
        out = bytearray()
        escaped = False

        for c in data:
            if escaped:
                byte = c ^ 0b00100000
                assert byte in RESERVED_BYTES
                out.append(byte)
                escaped = False
            elif c == Reserved.ESCAPE:
                escaped = True
            else:
                out.append(c)

        return out

    def data_received(self, data: bytes) -> None:
        _LOGGER.debug("Received data %s", data.hex())
        self._buffer.extend(data)

        if len(self._buffer) > MAX_BUFFER_SIZE:
            _LOGGER.debug(
                "Truncating buffer to %s bytes, it is growing too fast", MAX_BUFFER_SIZE
            )
            self._buffer = self._buffer[:MAX_BUFFER_SIZE]

        while self._buffer:
            if self._discarding_until_next_flag:
                if bytes([Reserved.FLAG]) not in self._buffer:
                    self._buffer.clear()
                    break

                self._discarding_until_next_flag = False
                _, _, self._buffer = self._buffer.partition(bytes([Reserved.FLAG]))

            try:
                # Find the index of the first reserved byte that isn't an escape byte
                reserved_index, reserved_byte = next(
                    (index, byte)
                    for index, byte in enumerate(self._buffer)
                    if byte in RESERVED_WITHOUT_ESCAPE
                )
            except StopIteration:
                break

            if reserved_byte == Reserved.FLAG:
                # Flag Byte marks the end of a frame
                frame_bytes = self._buffer[:reserved_index]
                self._buffer = self._buffer[reserved_index + 1 :]

                # Consecutive EOFs can be received, empty frames are ignored
                if not frame_bytes:
                    continue

                data = self._unstuff_bytes(frame_bytes)

                try:
                    frame = parse_frame(data)
                except Exception:
                    _LOGGER.debug(
                        "Failed to parse frame %r", frame_bytes, exc_info=True
                    )
                else:
                    self.frame_received(frame)
            elif reserved_byte == Reserved.CANCEL:
                _LOGGER.debug("Received cancel byte, clearing buffer")
                # All data received since the previous Flag Byte to be ignored
                self._buffer = self._buffer[reserved_index + 1 :]
            elif reserved_byte == Reserved.SUBSTITUTE:
                _LOGGER.debug("Received substitute byte, marking buffer as corrupted")
                # The data between the previous and the next Flag Byte is ignored
                self._discarding_until_next_flag = True
                self._buffer = self._buffer[reserved_index + 1 :]
            elif reserved_byte == Reserved.XON:
                # Resume transmission: not implemented!
                _LOGGER.debug("Received XON byte, resuming transmission")
                self._buffer.pop(reserved_index)
            elif reserved_byte == Reserved.XOFF:
                # Pause transmission: not implemented!
                _LOGGER.debug("Received XOFF byte, pausing transmission")
                self._buffer.pop(reserved_index)
            else:
                raise RuntimeError(
                    f"Unexpected reserved byte found: 0x{reserved_byte:02X}"
                )  # pragma: no cover

    def _handle_ack(self, frame: DataFrame | AckFrame) -> None:
        # Note that ackNum is the number of the next frame the receiver expects and it
        # is one greater than the last frame received.
        for ack_num_offset in range(-TX_K, 0):
            ack_num = (frame.ack_num + ack_num_offset) % 8
            fut = self._pending_data_frames.get(ack_num)

            if fut is None or fut.done():
                continue

            self._pending_data_frames[ack_num].set_result(True)

    def frame_received(self, frame: AshFrame) -> None:
        _LOGGER.debug("Received frame %r", frame)

        if isinstance(frame, DataFrame):
            self.data_frame_received(frame)
        elif isinstance(frame, RStackFrame):
            self.rstack_frame_received(frame)
        elif isinstance(frame, AckFrame):
            self.ack_frame_received(frame)
        elif isinstance(frame, NakFrame):
            self.nak_frame_received(frame)
        elif isinstance(frame, RstFrame):
            self.rst_frame_received(frame)
        elif isinstance(frame, ErrorFrame):
            self.error_frame_received(frame)
        else:
            raise TypeError(f"Unknown frame received: {frame}")  # pragma: no cover

    def data_frame_received(self, frame: DataFrame) -> None:
        # The Host may not piggyback acknowledgments and should promptly send an ACK
        # frame when it receives a DATA frame.
        if frame.frm_num == self._rx_seq:
            self._handle_ack(frame)
            self._rx_seq = (frame.frm_num + 1) % 8
            self._write_frame(AckFrame(res=0, ncp_ready=0, ack_num=self._rx_seq))

            self._ezsp_protocol.data_received(frame.ezsp_frame)
        elif frame.re_tx:
            # Retransmitted frames must be immediately ACKed even if they are out of
            # sequence
            self._write_frame(AckFrame(res=0, ncp_ready=0, ack_num=self._rx_seq))
        else:
            _LOGGER.debug("Received an out of sequence frame: %r", frame)
            self._write_frame(NakFrame(res=0, ncp_ready=0, ack_num=self._rx_seq))

    def rstack_frame_received(self, frame: RStackFrame) -> None:
        self._ncp_reset_code = None
        self._ncp_state = NcpState.CONNECTED

        self._tx_seq = 0
        self._rx_seq = 0
        self._change_ack_timeout(T_RX_ACK_INIT)
        self._ezsp_protocol.reset_received(frame.reset_code)

    def ack_frame_received(self, frame: AckFrame) -> None:
        self._handle_ack(frame)

    def nak_frame_received(self, frame: NakFrame) -> None:
        err = NotAcked(frame=frame)

        for fut in self._pending_data_frames.values():
            if not fut.done():
                fut.set_exception(err)

    def rst_frame_received(self, frame: RstFrame) -> None:
        self._ncp_reset_code = None
        self._ncp_state = NcpState.CONNECTED

    def error_frame_received(self, frame: ErrorFrame) -> None:
        _LOGGER.debug("NCP has entered failed state: %s", frame.reset_code)
        self._ncp_reset_code = frame.reset_code
        self._ncp_state = NcpState.FAILED

        # Cancel all pending requests
        self._enter_failed_state(self._ncp_reset_code)

    def _enter_failed_state(self, reset_code: t.NcpResetCode) -> None:
        exc = NcpFailure(code=reset_code)

        for fut in self._pending_data_frames.values():
            if not fut.done():
                fut.set_exception(exc)

        self._ezsp_protocol.reset_received(reset_code)

    def _write_frame(
        self,
        frame: AshFrame,
        *,
        prefix: tuple[Reserved] = (),
        suffix: tuple[Reserved] = (Reserved.FLAG,),
    ) -> None:
        if _LOGGER.isEnabledFor(logging.DEBUG):
            prefix_str = "".join([f"{r.name} + " for r in prefix])
            suffix_str = "".join([f" + {r.name}" for r in suffix])
            _LOGGER.debug("Sending frame %s%r%s", prefix_str, frame, suffix_str)

        data = bytes(prefix) + self._stuff_bytes(frame.to_bytes()) + bytes(suffix)
        _LOGGER.debug("Sending data  %s", data.hex())
        self._transport.write(data)

    def _change_ack_timeout(self, new_value: float) -> None:
        new_value = max(T_RX_ACK_MIN, min(new_value, T_RX_ACK_MAX))

        if abs(new_value - self._t_rx_ack) > 0.01:
            _LOGGER.debug(
                "Changing ACK timeout from %0.2f to %0.2f", self._t_rx_ack, new_value
            )

        self._t_rx_ack = new_value

    async def _send_data_frame(self, frame: AshFrame) -> None:
        if self._send_data_frame_semaphore.locked():
            _LOGGER.debug("Semaphore is locked, waiting")

        async with self._send_data_frame_semaphore:
            frm_num = None

            try:
                for attempt in range(ACK_TIMEOUTS):
                    if self._ncp_state == NcpState.FAILED:
                        _LOGGER.debug(
                            "NCP is in a failed state, not sending: %r", frame
                        )
                        raise NcpFailure(
                            t.NcpResetCode.ERROR_EXCEEDED_MAXIMUM_ACK_TIMEOUT_COUNT
                        )

                    if frm_num is None:
                        frm_num = self._tx_seq
                        self._tx_seq = (self._tx_seq + 1) % 8

                    # Use a fresh ACK number on every retry
                    frame = frame.replace(
                        frm_num=frm_num,
                        re_tx=(attempt > 0),
                        ack_num=self._rx_seq,
                    )

                    send_time = time.monotonic()

                    ack_future = asyncio.get_running_loop().create_future()
                    self._pending_data_frames[frm_num] = ack_future
                    self._write_frame(frame)

                    try:
                        async with asyncio_timeout(self._t_rx_ack):
                            await ack_future
                    except NotAcked:
                        _LOGGER.debug(
                            "NCP responded with NAK. Retrying (attempt %d)", attempt + 1
                        )

                        # For timing purposes, NAK can be treated as an ACK
                        delta = time.monotonic() - send_time
                        self._change_ack_timeout((7 / 8) * self._t_rx_ack + 0.5 * delta)

                        if attempt >= ACK_TIMEOUTS - 1:
                            self._enter_failed_state(
                                t.NcpResetCode.ERROR_EXCEEDED_MAXIMUM_ACK_TIMEOUT_COUNT
                            )
                            raise
                    except NcpFailure:
                        _LOGGER.debug(
                            "NCP has entered into a failed state, not retrying"
                        )
                        raise
                    except asyncio.TimeoutError:
                        _LOGGER.debug(
                            "No ACK received in %0.2fs (attempt %d)",
                            self._t_rx_ack,
                            attempt + 1,
                        )
                        # If a DATA frame acknowledgement is not received within the
                        # current timeout value, then t_rx_ack is doubled.
                        self._change_ack_timeout(2 * self._t_rx_ack)

                        if attempt >= ACK_TIMEOUTS - 1:
                            self._enter_failed_state(
                                t.NcpResetCode.ERROR_EXCEEDED_MAXIMUM_ACK_TIMEOUT_COUNT
                            )
                            raise
                    else:
                        # Whenever an acknowledgement is received, t_rx_ack is set to
                        # 7/8 of its current value plus 1/2 of the measured time for the
                        # acknowledgement.
                        delta = time.monotonic() - send_time
                        self._change_ack_timeout((7 / 8) * self._t_rx_ack + 0.5 * delta)

                        break
            finally:
                if frm_num is not None:
                    self._pending_data_frames.pop(frm_num)

    async def send_data(self, data: bytes) -> None:
        # Sending data is a critical operation and cannot really be cancelled
        await asyncio.shield(
            create_eager_task(
                self._send_data_frame(
                    # All of the other fields will be set during transmission/retries
                    DataFrame(frm_num=None, re_tx=None, ack_num=None, ezsp_frame=data)
                )
            )
        )

    def send_reset(self) -> None:
        self._write_frame(RstFrame(), prefix=(Reserved.CANCEL,))



import voluptuous as vol

from bellows.config import cv_uint16

from ..v4.config import EZSP_POLICIES_SHARED
from ..v9 import config as v9_config
from .types import EzspPolicyId

EZSP_SCHEMA = {
    **v9_config.EZSP_SCHEMA,
}

EZSP_POLICIES_SCH = {
    **EZSP_POLICIES_SHARED,
    **{vol.Optional(policy.name): cv_uint16 for policy in EzspPolicyId},
}



from zigpy.exceptions import APIException, ControllerException


class EzspError(APIException):
    pass


class InvalidCommandError(EzspError):
    pass


class ControllerError(ControllerException):
    pass


class StackAlreadyRunning(EzspError):
    pass



""""EZSP Protocol version 5 protocol handler."""
from __future__ import annotations

import logging

import voluptuous

import bellows.config
import bellows.types as t

from . import commands, config, types as v5_types
from ..v4 import EZSPv4

LOGGER = logging.getLogger(__name__)


class EZSPv5(EZSPv4):
    """EZSP Version 5 Protocol version handler."""

    VERSION = 5
    COMMANDS = commands.COMMANDS
    SCHEMAS = {
        bellows.config.CONF_EZSP_CONFIG: voluptuous.Schema(config.EZSP_SCHEMA),
        bellows.config.CONF_EZSP_POLICIES: voluptuous.Schema(config.EZSP_POLICIES_SCH),
    }
    types = v5_types

    def _ezsp_frame_tx(self, name: str) -> bytes:
        """Serialize the frame id."""
        cmd_id = self.COMMANDS[name][0]
        frame = [self._seq, 0x00, 0xFF, 0x00, cmd_id]
        return bytes(frame)

    def _ezsp_frame_rx(self, data: bytes) -> tuple[int, int, bytes]:
        """Handler for received data frame."""
        return data[0], data[4], data[5:]

    async def add_transient_link_key(
        self, ieee: t.EUI64, key: t.KeyData
    ) -> t.EmberStatus:
        (status,) = await self.addTransientLinkKey(ieee, key)
        return status

    async def pre_permit(self, time_s: int) -> None:
        """Add pre-shared TC Link key."""
        wild_card_ieee = v5_types.EUI64([0xFF] * 8)
        tc_link_key = v5_types.KeyData(b"ZigBeeAlliance09")
        await self.add_transient_link_key(wild_card_ieee, tc_link_key)



"""Protocol version 6 specific structs."""

import bellows.types.basic as basic
from bellows.types.struct import (  # noqa: F401
    EmberAesMmoHashContext,
    EmberApsFrame,
    EmberBindingTableEntry,
    EmberCurrentSecurityState,
    EmberGpAddress,
    EmberInitialSecurityState,
    EmberMulticastTableEntry,
    EmberNeighborTableEntry,
    EmberNetworkParameters,
    EmberRouteTableEntry,
    EmberTokTypeStackZllData,
    EmberTokTypeStackZllSecurity,
    EmberTransientKeyData,
    EmberZigbeeNetwork,
    EmberZllAddressAssignment,
    EmberZllDeviceInfoRecord,
    EmberZllInitialSecurityState,
    EmberZllNetwork,
    EmberZllSecurityAlgorithmData,
    EzspStruct,
)

from . import named


class EmberKeyStruct(EzspStruct):
    # A structure containing a key and its associated data.
    # A bitmask indicating the presence of data within the various fields
    # in the structure.
    bitmask: named.EmberKeyStructBitmask
    # The type of the key.
    type: named.EmberKeyType
    # The actual key data.
    key: named.KeyData
    # The outgoing frame counter associated with the key.
    outgoingFrameCounter: basic.uint32_t
    # The frame counter of the partner device associated with the key.
    incomingFrameCounter: basic.uint32_t
    # The sequence number associated with the key.
    sequenceNumber: basic.uint8_t
    # The IEEE address of the partner device also in possession of the key.
    partnerEUI64: named.EUI64


class EmberGpSinkListEntry(EzspStruct):
    # A sink list entry
    # The sink list type.
    type: basic.uint8_t
    # The EUI64 of the target sink.
    sinkEUI: named.EUI64
    # The short address of the target sink.
    sinkNodeId: named.EmberNodeId


class EmberGpProxyTableEntry(EzspStruct):
    """The internal representation of a proxy table entry."""

    # The link key to be used to secure this pairing link.
    securityLinkKey: named.KeyData
    # Internal status of the proxy table entry.
    status: named.EmberGpProxyTableEntryStatus
    # The tunneling options
    # (this contains both options and extendedOptions from the spec).
    options: basic.uint32_t
    # The addressing info of the GPD.
    gpd: EmberGpAddress
    # The assigned alias for the GPD.
    assignedAlias: named.EmberNodeId
    # The security options field.
    securityOptions: basic.uint8_t
    # The security frame counter of the GPD.
    gpdSecurityFrameCounter: named.EmberGpSecurityFrameCounter
    # The key to use for GPD.
    gpdKey: named.KeyData
    # The list of sinks (hardcoded to 2 which is the spec minimum).
    sinkList: basic.FixedList[EmberGpSinkListEntry, 2]
    # The groupcast radius.
    groupcastRadius: basic.uint8_t
    # The search counter
    searchCounter: basic.uint8_t


class EmberGpSinkTableEntry(EzspStruct):
    """The internal representation of a sink table entry."""

    # Internal status of the sink table entry
    status: named.EmberGpSinkTableEntryStatus
    # The tunneling options
    # (this contains both options and extendedOptions from the spec).
    options: basic.uint32_t
    # The addressing info of the GPD.
    gpd: EmberGpAddress
    # The device id for the GPD.
    deviceId: basic.uint8_t
    # The list of sinks (hardcoded to 2 which is the spec minimum).
    sinkList: basic.FixedList[EmberGpSinkListEntry, 2]
    # The assigned alias for the GPD.
    assignedAlias: named.EmberNodeId
    # The groupcast radius.
    groupcastRadius: basic.uint8_t
    # The security options field.
    securityOptions: basic.uint8_t
    # The security frame counter of the GPD.
    gpdSecurityFrameCounter: named.EmberGpSecurityFrameCounter
    # The key to use for GPD.
    gpdKey: named.KeyData


class EmberDutyCycleLimits(EzspStruct):
    """A structure containing duty cycle limit configurations.

    All limits are absolute, and are required to be as follows:
    suspLimit > critThresh > limitThresh
    For example:
    suspLimit = 250 (2.5%), critThresh = 180 (1.8%), limitThresh 100 (1.00%).
    """

    # The vendor identifier field shall contain the vendor identifier of the node.
    vendorId: basic.uint16_t
    # The vendor string field shall contain the vendor string of the node.
    vendorString: basic.FixedList[basic.uint8_t, 7]


class EmberPerDeviceDutyCycle(EzspStruct):
    """A structure containing per device overall duty cycle consumed

    up to the suspend limit).
    """

    # Node Id of device whose duty cycle is reported.
    nodeId: named.EmberNodeId
    # Amount of overall duty cycle consumed (up to suspend limit).
    dutyCycleConsumed: named.EmberDutyCycleHectoPct



import asyncio
import sys
import threading
from unittest import mock

if sys.version_info[:2] < (3, 11):
    from async_timeout import timeout as asyncio_timeout  # pragma: no cover
else:
    from asyncio import timeout as asyncio_timeout  # pragma: no cover

import pytest

from bellows.thread import EventLoopThread, ThreadsafeProxy


async def test_thread_start(monkeypatch):
    current_loop = asyncio.get_event_loop()
    loopmock = mock.MagicMock()

    monkeypatch.setattr(asyncio, "new_event_loop", lambda: loopmock)
    monkeypatch.setattr(asyncio, "set_event_loop", lambda loop: None)

    def mockrun(task):
        future = asyncio.run_coroutine_threadsafe(task, loop=current_loop)
        return future.result(1)

    loopmock.run_until_complete.side_effect = mockrun
    thread = EventLoopThread()
    thread_complete = await thread.start()
    await thread_complete

    assert loopmock.run_until_complete.call_count == 1
    assert loopmock.run_forever.call_count == 1
    assert loopmock.close.call_count == 1


class ExceptionCollector:
    def __init__(self):
        self.exceptions = []

    def __call__(self, thread_loop, context):
        exc = context.get("exception") or Exception(context["message"])
        self.exceptions.append(exc)


@pytest.fixture
async def thread():
    thread = EventLoopThread()
    await thread.start()
    thread.loop.call_soon_threadsafe(
        thread.loop.set_exception_handler, ExceptionCollector()
    )
    yield thread
    thread.force_stop()
    if thread.thread_complete is not None:
        async with asyncio_timeout(1):
            await thread.thread_complete
    [t.join(1) for t in threading.enumerate() if "bellows" in t.name]
    threads = [t for t in threading.enumerate() if "bellows" in t.name]
    assert len(threads) == 0


async def yield_other_thread(thread):
    await thread.run_coroutine_threadsafe(asyncio.sleep(0))

    exception_collector = thread.loop.get_exception_handler()
    if exception_collector.exceptions:
        raise exception_collector.exceptions[0]


async def test_thread_loop(thread):
    async def test_coroutine():
        return mock.sentinel.result

    future = asyncio.run_coroutine_threadsafe(test_coroutine(), loop=thread.loop)
    result = await asyncio.wrap_future(future, loop=asyncio.get_event_loop())
    assert result is mock.sentinel.result


async def test_thread_double_start(thread):
    previous_loop = thread.loop
    await thread.start()
    if sys.version_info[:2] >= (3, 6):
        threads = [t for t in threading.enumerate() if "bellows" in t.name]
        assert len(threads) == 1
    assert thread.loop is previous_loop


async def test_thread_already_stopped(thread):
    thread.force_stop()
    thread.force_stop()


async def test_thread_run_coroutine_threadsafe(thread):
    inner_loop = None

    async def test_coroutine():
        nonlocal inner_loop
        inner_loop = asyncio.get_event_loop()
        return mock.sentinel.result

    result = await thread.run_coroutine_threadsafe(test_coroutine())
    assert result is mock.sentinel.result
    assert inner_loop is thread.loop


async def test_proxy_callback(thread):
    obj = mock.MagicMock()
    proxy = ThreadsafeProxy(obj, thread.loop)
    obj.test.return_value = None
    proxy.test()
    await yield_other_thread(thread)
    assert obj.test.call_count == 1


async def test_proxy_async(thread):
    obj = mock.MagicMock()
    proxy = ThreadsafeProxy(obj, thread.loop)
    call_count = 0

    async def magic():
        nonlocal thread, call_count
        assert asyncio.get_event_loop() == thread.loop
        call_count += 1
        return mock.sentinel.result

    obj.test = magic
    result = await proxy.test()

    assert call_count == 1
    assert result == mock.sentinel.result


async def test_proxy_bad_function(thread):
    obj = mock.MagicMock()
    proxy = ThreadsafeProxy(obj, thread.loop)
    obj.test.return_value = mock.sentinel.value

    with pytest.raises(TypeError):
        proxy.test()
        await yield_other_thread(thread)


async def test_proxy_not_function():
    loop = asyncio.get_event_loop()
    obj = mock.MagicMock()
    proxy = ThreadsafeProxy(obj, loop)
    obj.test = mock.sentinel.value
    with pytest.raises(TypeError):
        proxy.test


async def test_proxy_no_thread():
    loop = asyncio.get_event_loop()
    obj = mock.MagicMock()
    proxy = ThreadsafeProxy(obj, loop)
    proxy.test()
    assert obj.test.call_count == 1


async def test_proxy_loop_closed():
    loop = asyncio.new_event_loop()
    obj = mock.MagicMock()
    proxy = ThreadsafeProxy(obj, loop)
    loop.close()
    proxy.test()
    assert obj.test.call_count == 0


async def test_thread_task_cancellation_after_stop(thread):
    loop = asyncio.get_event_loop()
    obj = mock.MagicMock()

    async def wait_forever():
        return await thread.loop.create_future()

    obj.wait_forever = wait_forever

    # Stop the thread while we're waiting
    loop.call_later(0.1, thread.force_stop)

    proxy = ThreadsafeProxy(obj, thread.loop)

    # The cancellation should propagate to the outer event loop
    with pytest.raises(asyncio.CancelledError):
        # This will stall forever without the patch
        async with asyncio_timeout(1):
            await proxy.wait_forever()



from __future__ import annotations

import asyncio
import logging
import random
from unittest.mock import MagicMock, call, patch

import pytest

from bellows import ash
import bellows.types as t


@pytest.fixture(autouse=True, scope="function")
def random_seed():
    random.seed(0)


class AshNcpProtocol(ash.AshProtocol):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.nak_state = False

    def frame_received(self, frame: ash.AshFrame) -> None:
        if self._ncp_reset_code is not None and not isinstance(frame, ash.RstFrame):
            ash._LOGGER.debug(
                "NCP in failure state %r, ignoring frame: %r",
                self._ncp_reset_code,
                frame,
            )
            self._write_frame(
                ash.ErrorFrame(version=2, reset_code=self._ncp_reset_code)
            )
            return

        if self.nak_state:
            asyncio.get_running_loop().call_later(
                2 * self._t_rx_ack,
                lambda: self._write_frame(
                    ash.NakFrame(res=0, ncp_ready=0, ack_num=self._rx_seq)
                ),
            )
            return

        super().frame_received(frame)

    def _enter_ncp_error_state(self, code: t.NcpResetCode | None) -> None:
        self._ncp_reset_code = code

        if code is None:
            self._ncp_state = ash.NcpState.CONNECTED
        else:
            self._ncp_state = ash.NcpState.FAILED

        ash._LOGGER.debug("Changing connectivity state: %r", self._ncp_state)
        ash._LOGGER.debug("Changing reset code: %r", self._ncp_reset_code)

        if self._ncp_state == ash.NcpState.FAILED:
            self._write_frame(
                ash.ErrorFrame(version=2, reset_code=self._ncp_reset_code)
            )

    def rst_frame_received(self, frame: ash.RstFrame) -> None:
        super().rst_frame_received(frame)

        self._tx_seq = 0
        self._rx_seq = 0
        self._change_ack_timeout(ash.T_RX_ACK_INIT)

        self._enter_ncp_error_state(None)
        self._write_frame(
            ash.RStackFrame(version=2, reset_code=t.NcpResetCode.RESET_SOFTWARE)
        )

    async def _send_data_frame(self, frame: ash.AshFrame) -> None:
        try:
            return await super()._send_data_frame(frame)
        except asyncio.TimeoutError:
            self._enter_ncp_error_state(
                t.NcpResetCode.ERROR_EXCEEDED_MAXIMUM_ACK_TIMEOUT_COUNT
            )
            raise

    def send_reset(self) -> None:
        raise NotImplementedError()


class FakeTransport:
    def __init__(self, receiver):
        self.receiver = receiver
        self.paused = False

    def write(self, data: bytes) -> None:
        if not self.paused:
            self.receiver.data_received(data)


class FakeTransportOneByteAtATime(FakeTransport):
    def write(self, data: bytes) -> None:
        for byte in data:
            super().write(bytes([byte]))


class FakeTransportRandomLoss(FakeTransport):
    def write(self, data: bytes) -> None:
        if random.random() < 0.20:
            return

        super().write(data)


class FakeTransportWithDelays(FakeTransport):
    def write(self, data):
        asyncio.get_running_loop().call_later(0, super().write, data)


def test_ash_exception_repr() -> None:
    assert (
        repr(ash.NotAcked(ash.NakFrame(res=0, ncp_ready=0, ack_num=1)))
        == "<NotAcked(frame=NakFrame(res=0, ncp_ready=0, ack_num=1))>"
    )
    assert (
        repr(ash.NcpFailure(t.NcpResetCode.RESET_SOFTWARE))
        == "<NcpFailure(code=<NcpResetCode.RESET_SOFTWARE: 11>)>"
    )


@pytest.mark.parametrize(
    "frame",
    [
        ash.RstFrame(),
        ash.AckFrame(res=0, ncp_ready=0, ack_num=1),
        ash.NakFrame(res=0, ncp_ready=0, ack_num=1),
        ash.RStackFrame(version=2, reset_code=t.NcpResetCode.RESET_SOFTWARE),
        ash.ErrorFrame(version=2, reset_code=t.NcpResetCode.RESET_SOFTWARE),
        ash.DataFrame(frm_num=0, re_tx=False, ack_num=1, ezsp_frame=b"test"),
    ],
)
def test_parse_frame(frame: ash.AshFrame) -> None:
    assert ash.parse_frame(frame.to_bytes()) == frame


def test_parse_frame_failure() -> None:
    with pytest.raises(ash.ParsingError):
        ash.parse_frame(b"test")


def test_ash_protocol_event_propagation() -> None:
    ezsp = MagicMock()
    protocol = ash.AshProtocol(ezsp)

    err = RuntimeError("test")
    protocol.connection_lost(err)
    assert ezsp.connection_lost.mock_calls == [call(err)]

    protocol.eof_received()
    assert ezsp.eof_received.mock_calls == [call()]


def test_stuffing():
    assert ash.AshProtocol._stuff_bytes(b"\x7E") == b"\x7D\x5E"
    assert ash.AshProtocol._stuff_bytes(b"\x11") == b"\x7D\x31"
    assert ash.AshProtocol._stuff_bytes(b"\x13") == b"\x7D\x33"
    assert ash.AshProtocol._stuff_bytes(b"\x18") == b"\x7D\x38"
    assert ash.AshProtocol._stuff_bytes(b"\x1A") == b"\x7D\x3A"
    assert ash.AshProtocol._stuff_bytes(b"\x7D") == b"\x7D\x5D"

    assert ash.AshProtocol._unstuff_bytes(b"\x7D\x5E") == b"\x7E"
    assert ash.AshProtocol._unstuff_bytes(b"\x7D\x31") == b"\x11"
    assert ash.AshProtocol._unstuff_bytes(b"\x7D\x33") == b"\x13"
    assert ash.AshProtocol._unstuff_bytes(b"\x7D\x38") == b"\x18"
    assert ash.AshProtocol._unstuff_bytes(b"\x7D\x3A") == b"\x1A"
    assert ash.AshProtocol._unstuff_bytes(b"\x7D\x5D") == b"\x7D"

    assert ash.AshProtocol._stuff_bytes(b"\x7F") == b"\x7F"
    assert ash.AshProtocol._unstuff_bytes(b"\x7F") == b"\x7F"


def test_pseudo_random_data_sequence():
    assert ash.PSEUDO_RANDOM_DATA_SEQUENCE.startswith(b"\x42\x21\xA8\x54\x2A")


def test_frame_parsing_errors():
    with pytest.raises(ash.ParsingError, match=r"Frame is too short:"):
        assert ash.RstFrame.from_bytes(b"\xC0\x38")

    with pytest.raises(ash.ParsingError, match=r"Invalid CRC bytes in frame"):
        assert ash.RstFrame.from_bytes(b"\xC0\xAB\xCD")


def test_rst_frame():
    assert ash.RstFrame() == ash.RstFrame()
    assert ash.RstFrame().to_bytes() == b"\xC0\x38\xBC"
    assert ash.RstFrame.from_bytes(b"\xC0\x38\xBC") == ash.RstFrame()
    assert str(ash.RstFrame()) == "RstFrame()"

    with pytest.raises(ash.ParsingError, match=r"Invalid data for RST frame:"):
        ash.RstFrame.from_bytes(ash.AshFrame.append_crc(b"\xC0\xAB"))


def test_rstack_frame():
    frm = ash.RStackFrame(version=2, reset_code=t.NcpResetCode.RESET_SOFTWARE)

    assert frm.to_bytes() == b"\xc1\x02\x0b\x0a\x52"
    assert ash.RStackFrame.from_bytes(frm.to_bytes()) == frm
    assert (
        str(frm)
        == "RStackFrame(version=2, reset_code=<NcpResetCode.RESET_SOFTWARE: 11>)"
    )

    with pytest.raises(
        ash.ParsingError, match=r"Invalid data length for RSTACK frame:"
    ):
        # Adding \xAB in the middle of the frame makes it invalid
        ash.RStackFrame.from_bytes(ash.AshFrame.append_crc(b"\xc1\x02\xab\x0b"))

    with pytest.raises(ash.ParsingError, match=r"Invalid version for RSTACK frame:"):
        # Version 3 is unknown
        ash.RStackFrame.from_bytes(ash.AshFrame.append_crc(b"\xc1\x03\x0b"))


def test_cancel_byte():
    ezsp = MagicMock()
    protocol = ash.AshProtocol(ezsp)
    protocol.frame_received = MagicMock(wraps=protocol.frame_received)

    protocol.data_received(bytes.fromhex("ddf9ff"))
    protocol.data_received(bytes.fromhex("1ac1020b0a527e"))  # starts with a CANCEL byte

    # We still parse out the RSTACK frame
    assert protocol.frame_received.mock_calls == [
        call(ash.RStackFrame(version=2, reset_code=t.NcpResetCode.RESET_SOFTWARE))
    ]
    assert not protocol._buffer

    # Parse byte-by-byte
    protocol.frame_received.reset_mock()

    for byte in bytes.fromhex("ddf9ff 1ac1020b0a527e"):
        protocol.data_received(bytes([byte]))

    # We still parse out the RSTACK frame
    assert protocol.frame_received.mock_calls == [
        call(ash.RStackFrame(version=2, reset_code=t.NcpResetCode.RESET_SOFTWARE))
    ]
    assert not protocol._buffer


def test_substitute_byte():
    ezsp = MagicMock()
    protocol = ash.AshProtocol(ezsp)
    protocol.frame_received = MagicMock(wraps=protocol.frame_received)

    protocol.data_received(bytes.fromhex("c0        38bc 7e"))  # RST frame
    assert protocol.frame_received.mock_calls == [call(ash.RstFrame())]
    protocol.data_received(bytes.fromhex("c0 18     38bc 7e"))  # RST frame + SUBSTITUTE
    assert protocol.frame_received.mock_calls == [call(ash.RstFrame())]  # ignored!
    protocol.data_received(bytes.fromhex("c1 020b   0a52 7e"))  # RSTACK frame
    assert protocol.frame_received.mock_calls == [
        call(ash.RstFrame()),
        call(ash.RStackFrame(version=2, reset_code=t.NcpResetCode.RESET_SOFTWARE)),
    ]

    assert not protocol._buffer


def test_xon_xoff_bytes():
    ezsp = MagicMock()
    protocol = ash.AshProtocol(ezsp)
    protocol.frame_received = MagicMock(wraps=protocol.frame_received)

    protocol.data_received(bytes.fromhex("c0 11 38bc 13 7e"))  # RST frame + XON + XOFF
    assert protocol.frame_received.mock_calls == [call(ash.RstFrame())]


def test_multiple_eof():
    ezsp = MagicMock()
    protocol = ash.AshProtocol(ezsp)
    protocol.frame_received = MagicMock(wraps=protocol.frame_received)

    protocol.data_received(bytes.fromhex("c0 38bc 7e 7e 7e"))  # RST frame
    assert protocol.frame_received.mock_calls == [call(ash.RstFrame())]


def test_discarding():
    ezsp = MagicMock()
    protocol = ash.AshProtocol(ezsp)
    protocol.frame_received = MagicMock(wraps=protocol.frame_received)

    # RST frame with embedded SUBSTITUTE
    protocol.data_received(bytes.fromhex("c0 18 38bc"))

    # Garbage: still ignored
    protocol.data_received(bytes.fromhex("aa bb cc dd ee ff"))

    # Frame boundary: we now will handle data
    protocol.data_received(bytes.fromhex("7e"))

    # Normal RST frame
    protocol.data_received(bytes.fromhex("c0    38bc 7e 7e 7e"))
    assert protocol.frame_received.mock_calls == [call(ash.RstFrame())]


def test_buffer_growth():
    ezsp = MagicMock()
    protocol = ash.AshProtocol(ezsp)

    # Receive a lot of bogus data
    for i in range(1000):
        protocol.data_received(b"\xEE" * 100)

    # Make sure our internal buffer doesn't blow up
    assert len(protocol._buffer) == ash.MAX_BUFFER_SIZE


async def test_sequence():
    loop = asyncio.get_running_loop()
    ezsp = MagicMock()
    transport = MagicMock()

    protocol = ash.AshProtocol(ezsp)
    protocol._write_frame = MagicMock(wraps=protocol._write_frame)
    protocol.connection_made(transport)

    # Normal send/receive
    loop.call_later(
        0,
        protocol.frame_received,
        ash.DataFrame(frm_num=0, re_tx=False, ack_num=1, ezsp_frame=b"rx 1"),
    )
    await protocol.send_data(b"tx 1")
    assert protocol._write_frame.mock_calls[-1] == call(
        ash.AckFrame(res=0, ncp_ready=0, ack_num=1)
    )

    assert protocol._rx_seq == 1
    assert protocol._tx_seq == 1
    assert ezsp.data_received.mock_calls == [call(b"rx 1")]

    # Skip ACK 2: we are out of sync!
    protocol.frame_received(
        ash.DataFrame(frm_num=2, re_tx=False, ack_num=1, ezsp_frame=b"out of sequence")
    )

    # We NAK it, it is out of sequence!
    assert protocol._write_frame.mock_calls[-1] == call(
        ash.NakFrame(res=0, ncp_ready=0, ack_num=1)
    )

    # Sequence numbers remain intact
    assert protocol._rx_seq == 1
    assert protocol._tx_seq == 1

    # Re-sync properly
    protocol.frame_received(
        ash.DataFrame(frm_num=1, re_tx=False, ack_num=1, ezsp_frame=b"rx 2")
    )

    assert ezsp.data_received.mock_calls == [call(b"rx 1"), call(b"rx 2")]

    # Trigger an error
    loop.call_later(
        0,
        protocol.frame_received,
        ash.ErrorFrame(version=2, reset_code=t.NcpResetCode.RESET_SOFTWARE),
    )

    with pytest.raises(ash.NcpFailure):
        await protocol.send_data(b"tx 2")


async def test_frame_parsing_failure_recovery(caplog) -> None:
    ezsp = MagicMock()
    protocol = ash.AshProtocol(ezsp)
    protocol.frame_received = MagicMock(spec_set=protocol.frame_received)

    protocol.data_received(
        ash.DataFrame(frm_num=0, re_tx=0, ack_num=0, ezsp_frame=b"frame 1").to_bytes()
        + bytes([ash.Reserved.FLAG])
    )

    with caplog.at_level(logging.DEBUG):
        protocol.data_received(
            ash.AshFrame.append_crc(b"\xFESome unknown frame")
            + bytes([ash.Reserved.FLAG])
        )

    assert "Some unknown frame" in caplog.text

    protocol.data_received(
        ash.DataFrame(frm_num=1, re_tx=0, ack_num=0, ezsp_frame=b"frame 2").to_bytes()
        + bytes([ash.Reserved.FLAG])
    )

    assert protocol.frame_received.mock_calls == [
        call(ash.DataFrame(frm_num=0, re_tx=0, ack_num=0, ezsp_frame=b"frame 1")),
        call(ash.DataFrame(frm_num=1, re_tx=0, ack_num=0, ezsp_frame=b"frame 2")),
    ]


async def test_ash_protocol_startup(caplog):
    """Simple EZSP startup: reset, version(4), then version(8)."""

    # We have branching dependent on `_LOGGER.isEnabledFor` so test it here
    caplog.set_level(logging.DEBUG)

    loop = asyncio.get_running_loop()

    ezsp = MagicMock()
    transport = MagicMock()

    protocol = ash.AshProtocol(ezsp)
    protocol._write_frame = MagicMock(wraps=protocol._write_frame)
    protocol.connection_made(transport)

    assert ezsp.connection_made.mock_calls == [call(protocol)]

    assert protocol._rx_seq == 0
    assert protocol._tx_seq == 0

    # ASH reset
    protocol.send_reset()
    loop.call_later(
        0,
        protocol.frame_received,
        ash.RStackFrame(version=2, reset_code=t.NcpResetCode.RESET_SOFTWARE),
    )

    await asyncio.sleep(0.01)

    assert ezsp.reset_received.mock_calls == [call(t.NcpResetCode.RESET_SOFTWARE)]
    assert protocol._write_frame.mock_calls == [
        call(ash.RstFrame(), prefix=(ash.Reserved.CANCEL,))
    ]

    protocol._write_frame.reset_mock()

    # EZSP version(4)
    loop.call_later(
        0,
        protocol.frame_received,
        ash.DataFrame(
            frm_num=0, re_tx=False, ack_num=1, ezsp_frame=b"\x00\x80\x00\x08\x02\x80g"
        ),
    )
    await protocol.send_data(b"\x00\x00\x00\x04")
    assert protocol._write_frame.mock_calls == [
        call(
            ash.DataFrame(
                frm_num=0, re_tx=False, ack_num=0, ezsp_frame=b"\x00\x00\x00\x04"
            )
        ),
        call(ash.AckFrame(res=0, ncp_ready=0, ack_num=1)),
    ]

    protocol._write_frame.reset_mock()

    # EZSP version(8)
    loop.call_later(
        0,
        protocol.frame_received,
        ash.DataFrame(
            frm_num=1,
            re_tx=False,
            ack_num=2,
            ezsp_frame=b"\x00\x80\x01\x00\x00\x08\x02\x80g",
        ),
    )
    await protocol.send_data(b"\x00\x00\x01\x00\x00\x08")
    assert protocol._write_frame.mock_calls == [
        call(
            ash.DataFrame(
                frm_num=1,
                re_tx=False,
                ack_num=1,
                ezsp_frame=b"\x00\x00\x01\x00\x00\x08",
            )
        ),
        call(ash.AckFrame(res=0, ncp_ready=0, ack_num=2)),
    ]


@patch("bellows.ash.T_RX_ACK_INIT", ash.T_RX_ACK_INIT / 100)
@patch("bellows.ash.T_RX_ACK_MIN", ash.T_RX_ACK_MIN / 100)
@patch("bellows.ash.T_RX_ACK_MAX", ash.T_RX_ACK_MAX / 100)
@pytest.mark.parametrize(
    "transport_cls",
    [
        FakeTransport,
        FakeTransportOneByteAtATime,
        FakeTransportRandomLoss,
        FakeTransportWithDelays,
    ],
)
async def test_ash_end_to_end(transport_cls: type[FakeTransport]) -> None:
    asyncio.get_running_loop()

    host_ezsp = MagicMock()
    ncp_ezsp = MagicMock()

    host = ash.AshProtocol(host_ezsp)
    ncp = AshNcpProtocol(ncp_ezsp)

    host_transport = transport_cls(ncp)
    ncp_transport = transport_cls(host)

    host.connection_made(host_transport)
    ncp.connection_made(ncp_transport)

    # Ping pong works
    await asyncio.gather(
        host.send_data(b"Hello 1!"),
        host.send_data(b"Hello 2!"),
    )
    assert ncp_ezsp.data_received.mock_calls == [call(b"Hello 1!"), call(b"Hello 2!")]

    await ncp.send_data(b"World!")
    assert host_ezsp.data_received.mock_calls == [call(b"World!")]

    ncp_ezsp.data_received.reset_mock()
    host_ezsp.data_received.reset_mock()

    # Let's pause the ncp so it can't ACK
    with patch.object(ncp_transport, "paused", True):
        send_task = asyncio.create_task(host.send_data(b"delayed"))
        await asyncio.sleep(host._t_rx_ack * 2)

    # It'll still succeed
    await send_task

    assert ncp_ezsp.data_received.mock_calls == [call(b"delayed")]

    ncp_ezsp.data_received.reset_mock()
    host_ezsp.data_received.reset_mock()

    # Let's let a request fail due to a connectivity issue
    with patch.object(ncp_transport, "paused", True):
        send_task = asyncio.create_task(host.send_data(b"host failure"))
        await asyncio.sleep(host._t_rx_ack * 15)

    with pytest.raises(asyncio.TimeoutError):
        await send_task

    ncp_ezsp.data_received.reset_mock()
    host_ezsp.data_received.reset_mock()

    # Simulate OOM on the NCP and send NAKs for a bit
    with patch.object(ncp, "nak_state", True):
        send_task = asyncio.create_task(host.send_data(b"ncp NAKing"))
        await asyncio.sleep(host._t_rx_ack)

    # It'll still succeed
    await send_task

    ncp_ezsp.data_received.reset_mock()
    host_ezsp.data_received.reset_mock()

    # When the NCP fails to receive a reply, it enters a failed state
    assert host._ncp_reset_code is None
    assert ncp._ncp_reset_code is None

    with patch.object(host_transport, "paused", True):
        send_task = asyncio.create_task(ncp.send_data(b"ncp failure"))
        await asyncio.sleep(ncp._t_rx_ack * 15)

    with pytest.raises(asyncio.TimeoutError):
        await send_task

    assert (
        host._ncp_reset_code is t.NcpResetCode.ERROR_EXCEEDED_MAXIMUM_ACK_TIMEOUT_COUNT
    )
    assert (
        ncp._ncp_reset_code is t.NcpResetCode.ERROR_EXCEEDED_MAXIMUM_ACK_TIMEOUT_COUNT
    )

    ncp_ezsp.data_received.reset_mock()
    host_ezsp.data_received.reset_mock()

    # All communication attempts with it will fail until it is reset
    with pytest.raises(ash.NcpFailure):
        await host.send_data(b"test")

    host.send_reset()
    await asyncio.sleep(0.01)
    await host.send_data(b"test")

    # Trigger a failure caused by excessive NAKs
    ncp._t_rx_ack = ash.T_RX_ACK_INIT / 1000
    host._t_rx_ack = ash.T_RX_ACK_INIT / 1000

    with patch.object(ncp, "nak_state", True):
        with pytest.raises(ash.NotAcked):
            await host.send_data(b"ncp NAKing until failure")



from bellows.ezsp.v12.types.named import *  # noqa: F401, F403
import bellows.types as t


class sl_zb_sec_man_derived_key_type_t(t.enum16):
    """Derived keys are calculated when performing Zigbee crypto operations.
    The stack makes use of these derivations.
    """

    # Perform no derivation; use the key as is.
    NONE = 0
    # Perform the Key-Transport-Key hash.
    KEY_TRANSPORT_KEY = 1
    # Perform the Key-Load-Key hash.
    KEY_LOAD_KEY = 2
    # Perform the Verify Key hash.
    VERIFY_KEY = 3
    # Perform a simple AES hash of the key for TC backup.
    TC_SWAP_OUT_KEY = 4
    # For a TC using hashed link keys, hashed the root key against the supplied EUI in
    # context.
    TC_HASHED_LINK_KEY = 5


class EzspValueId(t.enum8):
    # Identifies a value.

    # The contents of the node data stack token.
    VALUE_TOKEN_STACK_NODE_DATA = 0x00
    # The types of MAC passthrough messages that the host wishes to receive.
    VALUE_MAC_PASSTHROUGH_FLAGS = 0x01
    # The source address used to filter legacy EmberNet messages when the
    # MAC_PASSTHROUGH_EMBERNET_SOURCE flag is set in VALUE_MAC_PASSTHROUGH_FLAGS.
    VALUE_EMBERNET_PASSTHROUGH_SOURCE_ADDRESS = 0x02
    # The number of available message buffers.
    VALUE_FREE_BUFFERS = 0x03
    # Selects sending synchronous callbacks in ezsp-uart.
    VALUE_UART_SYNCH_CALLBACKS = 0x04
    # The maximum incoming transfer size for the local node.
    VALUE_MAXIMUM_INCOMING_TRANSFER_SIZE = 0x05
    # The maximum outgoing transfer size for the local node.
    VALUE_MAXIMUM_OUTGOING_TRANSFER_SIZE = 0x06
    # A boolean indicating whether stack tokens are written to persistent
    # storage as they change.
    VALUE_STACK_TOKEN_WRITING = 0x07
    # A read-only value indicating whether the stack is currently performing a rejoin.
    VALUE_STACK_IS_PERFORMING_REJOIN = 0x08
    # A list of EmberMacFilterMatchData values.
    VALUE_MAC_FILTER_LIST = 0x09
    # The Ember Extended Security Bitmask.
    VALUE_EXTENDED_SECURITY_BITMASK = 0x0A
    # The node short ID.
    VALUE_NODE_SHORT_ID = 0x0B
    # The descriptor capability of the local node.
    VALUE_DESCRIPTOR_CAPABILITY = 0x0C
    # The stack device request sequence number of the local node.
    VALUE_STACK_DEVICE_REQUEST_SEQUENCE_NUMBER = 0x0D
    # Enable or disable radio hold-off.
    VALUE_RADIO_HOLD_OFF = 0x0E
    # The flags field associated with the endpoint data.
    VALUE_ENDPOINT_FLAGS = 0x0F
    # Enable/disable the Mfg security config key settings.
    VALUE_MFG_SECURITY_CONFIG = 0x10
    # Retrieves the version information from the stack on the NCP.
    VALUE_VERSION_INFO = 0x11
    # This will get/set the rejoin reason noted by the host for a subsequent call to
    # emberFindAndRejoinNetwork(). After a call to emberFindAndRejoinNetwork() the
    # host's rejoin reason will be set to REJOIN_REASON_NONE. The NCP will store the
    # rejoin reason used by the call to emberFindAndRejoinNetwork()
    VALUE_NEXT_HOST_REJOIN_REASON = 0x12
    # This is the reason that the last rejoin took place. This value may only be
    # retrieved, not set. The rejoin may have been initiated by the stack (NCP) or the
    # application (host). If a host initiated a rejoin the reason will be set by default
    # to REJOIN_DUE_TO_APP_EVENT_1. If the application wishes to denote its own rejoin
    # reasons it can do so by calling ezspSetValue(VALUE_HOST_REJOIN_REASON,
    # REJOIN_DUE_TO_APP_EVENT_X). X is a number corresponding to one of the app events
    # defined. If the NCP initiated a rejoin it will record this value internally for
    # retrieval by ezspGetValue(VALUE_REAL_REJOIN_REASON).
    VALUE_LAST_REJOIN_REASON = 0x13
    # The next ZigBee sequence number.
    VALUE_NEXT_ZIGBEE_SEQUENCE_NUMBER = 0x14
    # CCA energy detect threshold for radio.
    VALUE_CCA_THRESHOLD = 0x15
    # The threshold value for a counter
    VALUE_SET_COUNTER_THRESHOLD = 0x17
    # Resets all counters thresholds to 0xFF
    VALUE_RESET_COUNTER_THRESHOLDS = 0x18
    # Clears all the counters
    VALUE_CLEAR_COUNTERS = 0x19
    # The node's new certificate signed by the CA.
    EZSP_VALUE_CERTIFICATE_283K1 = 0x1A
    # The Certificate Authority's public key.
    EZSP_VALUE_PUBLIC_KEY_283K1 = 0x1B
    # The node's new static private key.
    EZSP_VALUE_PRIVATE_KEY_283K1 = 0x1C
    # The NWK layer security frame counter value
    VALUE_NWK_FRAME_COUNTER = 0x23
    # The APS layer security frame counter value
    VALUE_APS_FRAME_COUNTER = 0x24
    # Sets the device type to use on the next rejoin using device type
    VALUE_RETRY_DEVICE_TYPE = 0x25
    # Setting this byte enables R21 behavior on the NCP.
    VALUE_ENABLE_R21_BEHAVIOR = 0x29
    # Configure the antenna mode(0-primary,1-secondary,2- toggle on tx ack fail).
    VALUE_ANTENNA_MODE = 0x30
    # Enable or disable packet traffic arbitration.
    VALUE_ENABLE_PTA = 0x31
    # Set packet traffic arbitration configuration options.
    VALUE_PTA_OPTIONS = 0x32
    # Configure manufacturing library options(0-non-CSMA transmits,1-CSMA transmits).
    VALUE_MFGLIB_OPTIONS = 0x33
    # Sets the flag to use either negotiated power by link power delta (LPD) or fixed
    # power value provided by user while forming/joining a network for packet
    # transmissions on subghz interface. This is mainly for testing purposes.
    VALUE_USE_NEGOTIATED_POWER_BY_LPD = 0x34
    # Set packet traffic arbitration configuration PWM options.
    VALUE_PTA_PWM_OPTIONS = 0x35
    # Set packet traffic arbitration directional priority pulse width in microseconds.
    VALUE_PTA_DIRECTIONAL_PRIORITY_PULSE_WIDTH = 0x36
    # Set packet traffic arbitration phy select timeout(ms)
    VALUE_PTA_PHY_SELECT_TIMEOUT = 0x37
    # Configure the RX antenna mode: (0-do not switch; 1- primary; 2-secondary;
    # 3-RX antenna diversity).
    VALUE_ANTENNA_RX_MODE = 0x38
    # Configure the timeout to wait for the network key before failing a join.
    VALUE_NWK_KEY_TIMEOUT = 0x39
    # The number of failed CSMA attempts due to failed CCA made by the MAC before
    # continuing transmission with CCA disabled. This is the same as calling the
    # emberForceTxAfterFailedCca(uint8_t csmaAttempts) API. A value of 0 disables the
    # feature
    VALUE_FORCE_TX_AFTER_FAILED_CCA_ATTEMPTS = 0x3A
    # The length of time, in seconds, that a trust center will store a transient link
    # key that a device can use to join its network. A transient key is added with a
    # call to emberAddTransientLinkKey. After the transient key is added, it will be
    # removed once this amount of time has passed. A joining device will not be able to
    # use that key to join until it is added again on the trust center. The default
    # value is 300 seconds (5 minutes).
    VALUE_TRANSIENT_KEY_TIMEOUT_S = 0x3B
    # Cumulative energy usage metric since the last value reset of the coulomb counter
    # plugin. Setting this value will reset the coulomb counter.
    VALUE_COULOMB_COUNTER_USAGE = 0x3C
    # When scanning, configure the maximum number of beacons to store in cache. Each
    # beacon consumes one packet buffer in RAM.
    VALUE_MAX_BEACONS_TO_STORE = 0x3D
    # Set the mask to filter out unacceptable child timeout options on a router.
    VALUE_END_DEVICE_TIMEOUT_OPTIONS_MASK = 0x3E
    # The end device keep alive mode supported by the parent.
    VALUE_END_DEVICE_KEEP_ALIVE_SUPPORT_MODE = 0x3F
    # Return the active radio config.
    VALUE_ACTIVE_RADIO_CONFIG = 0x41
    # Timeout in milliseconds to store entries in the transient device table.
    # If the devices are not authenticated before the timeout, the entry shall be
    # purged.
    VALUE_TRANSIENT_DEVICE_TIMEOUT = 0x43
    # Return information about the key storage on an NCP.
    # Returns 0 if keys are in classic key storage, and 1 if they
    # are located in PSA key storage. Read only.
    VALUE_KEY_STORAGE_VERSION = 0x44
    # Return activation state about TC Delayed Join on an NCP.  A return value of
    # 0 indicates that the feature is not activated.
    VALUE_DELAYED_JOIN_ACTIVATION = 0x45



from __future__ import annotations

import logging

import zigpy.device
import zigpy.endpoint
import zigpy.profiles.zgp
import zigpy.profiles.zha
import zigpy.profiles.zll
import zigpy.zdo.types as zdo_t

import bellows.types as t

LOGGER = logging.getLogger(__name__)

PROFILE_TO_DEVICE_TYPE = {
    zigpy.profiles.zha.PROFILE_ID: zigpy.profiles.zha.DeviceType,
    zigpy.profiles.zll.PROFILE_ID: zigpy.profiles.zll.DeviceType,
    zigpy.profiles.zgp.PROFILE_ID: zigpy.profiles.zgp.DeviceType,
}


class EZSPEndpoint(zigpy.endpoint.Endpoint):
    def __init__(
        self,
        device: zigpy.device.Device,
        endpoint_id: int,
        descriptor: zdo_t.SimpleDescriptor,
    ) -> None:
        super().__init__(device, endpoint_id)

        self.profile_id = descriptor.profile

        if self.profile_id in PROFILE_TO_DEVICE_TYPE:
            self.device_type = PROFILE_TO_DEVICE_TYPE[self.profile_id](
                descriptor.device_type
            )
        else:
            self.device_type = descriptor.device_type

        for cluster in descriptor.input_clusters:
            self.add_input_cluster(cluster)

        for cluster in descriptor.output_clusters:
            self.add_output_cluster(cluster)

        self.status = zigpy.endpoint.Status.ZDO_INIT

    @property
    def manufacturer(self) -> str:
        """Manufacturer."""
        return "Silicon Labs"

    @property
    def model(self) -> str:
        """Model."""
        return "EZSP"

    async def add_to_group(self, grp_id: int, name: str = None) -> t.EmberStatus:
        if grp_id in self.member_of:
            return t.EmberStatus.SUCCESS

        app = self.device.application
        status = await app.multicast.subscribe(grp_id)
        if status != t.EmberStatus.SUCCESS:
            self.debug("Couldn't subscribe to 0x%04x group", grp_id)
            return status

        group = app.groups.add_group(grp_id, name)
        group.add_member(self)
        return status

    async def remove_from_group(self, grp_id: int) -> t.EmberStatus:
        if grp_id not in self.member_of:
            return t.EmberStatus.SUCCESS

        app = self.device.application
        status = await app.multicast.unsubscribe(grp_id)
        if status != t.EmberStatus.SUCCESS:
            self.debug("Couldn't unsubscribe 0x%04x group", grp_id)
            return status

        app.groups[grp_id].remove_member(self)
        return status
